import logging

from ..ad_logger import PasswordRemovalFormatter


def test_password_removal(caplog):
    logger = logging.getLogger("testing")
    password = "hunter2"
    settings = {"global": {"password": password}}
    formatter = PasswordRemovalFormatter(settings=settings)

    # Capture logs on the "testing" logger at level DEBUG or higher
    caplog.set_level(logging.DEBUG, "testing")
    # Patch pytest logging to use our custom log formatter
    caplog.handler.setFormatter(formatter)

    logger.debug("password: %r", password)
    assert password not in caplog.text



import copy
import json
import uuid
from contextlib import ExitStack
from unittest.mock import MagicMock
from unittest.mock import Mock
from unittest.mock import patch
from uuid import uuid4

import requests
from os2mo_helpers.mora_helpers import MoraHelper
from winrm import Session

from ..ad_common import AD
from ..ad_writer import MORESTSource
from .test_utils import TestADWriterMixin

MO_ROOT_ORG_UNIT_UUID = "not-a-mo-org-unit-uuid"
MO_CHILD_ORG_UNIT_UUID = uuid4()
MO_UUID = "not-a-uuid"
AD_UUID_FIELD = "uuidField"
UNKNOWN_CPR_NO = "not-a-cpr-no"
MO_AD_IT_SYSTEM_UUID = uuid4()


class MockAD(AD):
    def __init__(self):
        self.session = Mock(spec=Session)
        self.all_settings = {"primary": {"search_base": ""}}


class MockADParameterReader(TestADWriterMixin):
    """Mock implementation of `ADParameterReader` which always returns the same
    AD user."""

    generate_dynamic_person = False

    def read_user(self, cpr=None, **kwargs):
        return self._prepare_get_from_ad(ad_transformer=None)

    def read_it_all(self, **kwargs):
        return [self.read_user()]

    def cache_all(self, **kwargs):
        return self.read_it_all()

    def get_all_samaccountname_values(self):
        return {self.read_user()["SamAccountName"]}


class MockEmptyADReader(MockADParameterReader):
    """Mock implementation of `ADParameterReader` which simulates an empty AD"""

    def read_user(self, **kwargs):
        return None

    def read_it_all(self, **kwargs):
        return []

    def cache_all(self, **kwargs):
        return self.read_it_all()


class MockUnknownCPRADParameterReader(MockADParameterReader):
    def read_user(self, cpr=None, **kwargs):
        if cpr == UNKNOWN_CPR_NO:
            return None
        return super().read_user(cpr=cpr, **kwargs)


class MockMORESTSource(MORESTSource):
    def __init__(self, from_date, to_date):
        self.from_date = from_date
        self.to_date = to_date

    def get_engagement_dates(self, uuid):
        # Return 2-tuple of (from_dates, to_dates)
        return [self.from_date], [self.to_date]


class MockMORESTSourcePreview(MORESTSource):
    def __init__(self):
        self.helper = MockMoraHelper("cpr")

    def find_primary_engagement(self, uuid):
        return "employment-number", "title", "eng-org-unit", "eng-uuid"

    def read_user(self, uuid):
        mo_user = self.helper.read_user()
        mo_user.update(
            {
                "givenname": "Tester",
                "surname": "Testesen",
            }
        )
        return mo_user

    def get_engagement_dates(self, uuid):
        return [None], [None]


class MockLoraCache:
    # This implements enough of the real `LoraCache` to make
    # `ad_sync.AdMoSync._edit_engagement` happy.

    def __init__(self, mo_values, mo_engagements=None):
        self._mo_values = mo_values
        self._mo_engagements = mo_engagements

    @property
    def users(self):
        return {self._mo_values["uuid"]: [self._mo_values]}

    @property
    def engagements(self):
        extensions = {"udvidelse_%d" % n: "old mo value #%d" % n for n in range(1, 11)}
        if self._mo_engagements:
            return {
                eng["uuid"]: [
                    {
                        "uuid": eng["uuid"],
                        "user": self._mo_values["uuid"],
                        "unit": MO_ROOT_ORG_UNIT_UUID,
                        "user_key": "engagement_%s" % self._mo_values["uuid"],
                        "primary_boolean": eng["is_primary"],
                        "from_date": eng["validity"]["from"],
                        "to_date": eng["validity"]["to"],
                        "extensions": extensions,
                        "job_function": eng.get("job_function", {}).get("uuid"),
                        "primary_type": None,
                        "engagement_type": None,
                    }
                ]
                for eng in self._mo_engagements
            }
        else:
            return {
                "engagement_uuid": [
                    {
                        "uuid": "engagement_uuid",
                        "user": self._mo_values["uuid"],
                        "unit": MO_ROOT_ORG_UNIT_UUID,
                        "user_key": "engagement_%s" % self._mo_values["uuid"],
                        "primary_boolean": True,
                        "from_date": "1960-01-01",
                        "to_date": None,
                        "extensions": extensions,
                        # Additional keys read by `AdLifeCycle._gen_filtered_employees`
                        "job_function": None,
                        "primary_type": None,
                        "engagement_type": None,
                    }
                ]
            }


class MockLoraCacheExtended(MockLoraCache):
    """Mocks enough of `LoraCache` to test `AdLifeCycle`"""

    def populate_cache(self, **kwargs):
        pass

    def calculate_derived_unit_data(self):
        pass

    def calculate_primary_engagements(self):
        pass

    @property
    def units(self):
        # Return a single org unit (= the root org unit)
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                }
            ]
        }

    @property
    def classes(self):
        return {None: {}}

    def _load_settings(self):
        return {}

    def _read_org_uuid(self):
        return "not-a-mo-org-uuid"


class MockLoraCacheEmptyEmployee(MockLoraCacheExtended):
    @property
    def users(self):
        return {self._mo_values["uuid"]: []}


class MockLoraCacheEmptyUnit(MockLoraCacheExtended):
    """Mock a LoraCache where there are no organisational units"""

    @property
    def units(self):
        return {}


class MockLoraCacheDanglingParentUnit(MockLoraCacheExtended):
    """Mock a LoraCache where organisational unit we look for has an unknown
    parent organisational unit UUID.
    """

    @property
    def units(self):
        return {
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": uuid4(),
                }
            ],
        }


class MockLoraCacheParentChildUnit(MockLoraCacheExtended):
    """Mock a LoraCache where a child unit points correctly to its parent unit
    (which is also the root unit in this case.)
    """

    @property
    def units(self):
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": MO_ROOT_ORG_UNIT_UUID,
                }
            ],
        }


class MockLoraCacheParentUnitUnset(MockLoraCacheExtended):
    """Mock a LoraCache where a child unit does not point correctly to its
    parent unit, due to its 'parent' key being None.
    """

    @property
    def units(self):
        return {
            MO_ROOT_ORG_UNIT_UUID: [
                {
                    "uuid": MO_ROOT_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
            MO_CHILD_ORG_UNIT_UUID: [
                {
                    "uuid": MO_CHILD_ORG_UNIT_UUID,
                    "parent": None,
                }
            ],
        }


class MockLoraCacheUnitAddress(MockLoraCacheExtended):

    _job_function_class_uuid = "job_function_class_uuid"

    def __init__(self, address_value):
        self._address_value = address_value
        mo_values = {"uuid": MO_UUID}
        mo_engagements = [
            {
                "uuid": "eng_%s" % MO_UUID,
                "is_primary": True,
                "validity": {"from": "1960-01-01", "to": None},
                "job_function": {"uuid": self._job_function_class_uuid},
            }
        ]
        super().__init__(mo_values, mo_engagements=mo_engagements)

    @property
    def classes(self):
        return {self._job_function_class_uuid: {"title": "Job function title"}}

    @property
    def addresses(self):
        address_uuid = "address-uuid"
        address = {
            "unit": MO_ROOT_ORG_UNIT_UUID,
            "scope": "DAR",
            "value": self._address_value,
        }
        return {address_uuid: [address]}


class MockMoraHelper(MoraHelper):
    def __init__(self, cpr, mo_uuid=None, read_ou_addresses=None):
        self._mo_user = {"cpr_no": cpr, "uuid": mo_uuid or MO_UUID}
        self._read_ou_addresses = read_ou_addresses or {}
        self._read_user_calls = []
        self._mo_post_calls = []
        super().__init__()

    def read_organisation(self):
        return "not-a-org-uuid"

    def read_user(self, user_cpr=None, **kwargs):
        self._read_user_calls.append(user_cpr)
        if user_cpr == UNKNOWN_CPR_NO:
            return None
        return self._mo_user

    def read_all_users(self):
        return [self._mo_user]

    def read_ou(self, uuid):
        return {
            "name": "org_unit_name",
            "user_key": "org_unit_user_key",
            "org_unit_type": {"uuid": "org_unit_type_uuid"},
            "org_unit_level": {"uuid": "org_unit_level_uuid"},
            "parent": None,
        }

    def read_ou_address(self, uuid, scope=None, **kwargs):
        if scope == "EMAIL":
            return []
        if scope == "DAR":
            return self._read_ou_addresses

    def read_engagement_manager(self, engagement_uuid):
        return {}

    def get_e_addresses(self, e_uuid, scope=None):
        return [
            {
                "value": "address-value",
                "address_type": {"uuid": "address-type-uuid"},
            }
        ]

    def _mo_post(self, url, payload, force=True):
        # Raise exception if payload cannot be serialized as JSON, e.g. if it contains
        # types that cannot be serialized as JSON (`uuid.UUID`, etc.)
        json.dumps(payload)
        # Record the MO API call
        self._mo_post_calls.append(dict(url=url, payload=payload, force=force))
        # Mock the MO API response
        mock_response = Mock(spec=requests.Response, status_code=201)
        mock_response.json.return_value = str(uuid.uuid4())
        return mock_response


class MockADWriterContext(ExitStack):
    """Mock enough of `ADWriter` dependencies to allow it to instantiate in a test.
    Usage:
    >>> with MockADWriterContext():
    >>>     ad_writer = ADWriter(...)
    >>>     ad_writer.some_method(...)
    """

    all_settings = {
        "primary": {
            "method": "ntlm",
            "cpr_separator": "",
            "system_user": "system_user",
            "password": "password",
            "search_base": "search_base",
            "cpr_field": "cpr_field",  # read by `ADWriter.get_from_ad`
            "properties": [],  # read by `ADWriter._properties`
        },
        "primary_write": {
            "cpr_field": "cpr_field",
            "uuid_field": "uuid_field",
            "org_field": "org_field",
            "level2orgunit_field": "level2orgunit_field",
            "level2orgunit_type": "level2orgunit_type",
            "upn_end": "upn_end",
            "mo_to_ad_fields": {},
            "template_to_ad_fields": {},
        },
        "global": {"mora.base": "", "servers": ["server"]},
    }

    def __init__(self, **kwargs):
        super().__init__()
        settings = copy.deepcopy(self.all_settings)
        template_to_ad_fields = kwargs.get("template_to_ad_fields", {})
        self._read_ou_addresses = kwargs.get("read_ou_addresses")
        settings["primary_write"]["template_to_ad_fields"].update(template_to_ad_fields)
        self._settings = settings

    def __enter__(self):
        super().__enter__()
        for ctx in self._context_managers:
            self.enter_context(ctx)
        return self

    @property
    def _context_managers(self):
        prefix = "integrations.ad_integration"
        yield patch(f"{prefix}.ad_common.read_settings", return_value=self._settings)
        yield patch(
            f"{prefix}.ad_writer.ADWriter._create_session", return_value=MagicMock()
        )
        yield patch(
            f"{prefix}.ad_writer.MORESTSource", return_value=MockMORESTSourcePreview()
        )
        yield patch(
            f"{prefix}.user_names.ADParameterReader",
            return_value=MockADParameterReader(),
        )
        yield patch(
            f"{prefix}.ad_writer.MoraHelper",
            return_value=MockMoraHelper(
                cpr="",
                read_ou_addresses=self._read_ou_addresses,
            ),
        )



import logging
from typing import Any
from typing import OrderedDict

from .config import get_changed_at_settings

logger = logging.getLogger("sdChangedAt")


def cpr_env_filter(entity: OrderedDict[str, Any]) -> bool:
    settings = get_changed_at_settings()

    cpr = entity["PersonCivilRegistrationIdentifier"]

    if settings.sd_exclude_cprs_mode:
        # The CPRs in the sd_cprs should be excluded from further processing
        # and all other CPRs should be processed
        process_cpr = cpr not in settings.sd_cprs
    else:
        # In this case "exclude mode" is False, i.e. "include mode" is True
        # which means that the cprs in sd_cprs are the ONLY ones that should
        # be processed further
        process_cpr = cpr in settings.sd_cprs

    if not process_cpr and settings.sd_exclude_cprs_mode:
        logger.warning(f"*** SKIPPING employee with cpr={cpr} ***")

    return process_cpr



from asyncio import run, sleep
from functools import partial
from pathlib import Path
from typing import Dict, Iterable, Optional
from uuid import UUID

from more_itertools import flatten

from integrations.gir.initial_read.emd_csv import (
    AddressKlasses,
    UUIDGenerator,
    gen_single_klass,
    read_emd,
)
from integrations.gir.initial_read.mo_lora_client import MoLoRaClient
from integrations.gir.initial_read.org_unit_csv import PartialManager, read_csv
from os2mo_data_import.Clients.LoRa.default_mo import ValidMo
from os2mo_data_import.Clients.LoRa.model import Klasse
from os2mo_data_import.Clients.MO.model import Engagement, Manager, OrgUnit
from os2mo_data_import.Clients.MO.model_parts.interface import MoObj
from os2mo_data_import.util import generate_uuid


def missing_ous(outu, oulu):
    """
    We are missing some org units. We create dummies to move along
    :param outu:
    :param oulu:
    :return:
    """

    def ou(user_key):
        """
        convenience
        :param user_key:
        :return:
        """
        return [
            OrgUnit.from_simplified_fields(
                uuid=generate_uuid("ou_any_seed_yo" + user_key),
                user_key=user_key,
                name="MISSING INFO",
                org_unit_type_uuid=outu,
                org_unit_level_uuid=oulu,
            )
        ]

    missing_department_nos = [
        "0",
        "50001401",
        "50003553",
        "50002830",
        "50050843",
        "50057275",
        "50003724",
        "50057678",
        "50000706",
        "50001633",
        "50002336",
        "50053954",
        "50054978",
    ]
    return list(map(ou, missing_department_nos))


def klasses_from_uuid_gen(
    facet_uuid: UUID, organisation_uuid: UUID, uuid_gen: UUIDGenerator
):
    def klasse_gen(uuid, user_key):
        return Klasse.from_simplified_fields(
            facet_uuid=facet_uuid,
            uuid=uuid,
            user_key=user_key,
            scope=None,
            organisation_uuid=organisation_uuid,
            title=user_key,
        )

    return [
        klasse_gen(uuid, user_key) for user_key, uuid in uuid_gen.get_all().items()
    ]


def gen_managers(
    partial_managers: Iterable[PartialManager],
    eng_mapping: Dict[str, Engagement],
    responsibility_uuid: UUID,
    manager_level_uuid: UUID,
    manager_type_uuid: UUID,
) -> Iterable[Manager]:
    def manager_gen(partial_manager: PartialManager) -> Optional[Manager]:
        try:
            eng = eng_mapping[partial_manager.engagement_user_key]
        except KeyError:
            return None

        return Manager.from_simplified_fields(
            uuid=generate_uuid(
                "manager"
                + partial_manager.engagement_user_key
                + str(partial_manager.org_unit_uuid)
            ),
            org_unit_uuid=partial_manager.org_unit_uuid,
            person_uuid=eng.person.uuid,
            responsibility_uuid=responsibility_uuid,
            manager_level_uuid=manager_level_uuid,
            manager_type_uuid=manager_type_uuid,
            from_date=eng.validity.from_date,
            to_date=eng.validity.to_date,
        )

    filtered_pm = filter(lambda x: x.engagement_user_key != "", partial_managers)
    return list(filter(lambda x: x is not None, map(manager_gen, filtered_pm)))


def read_values(base_path: Path) -> Iterable[Iterable[MoObj]]:
    base = ValidMo.from_scratch()
    gen_org_klass = partial(
        gen_single_klass, organisation_uuid=base.organisation.uuid
    )
    base_objs = [x[1] for x in base]
    address_classes = AddressKlasses.from_simplified_fields(
        organisation_uuid=base.organisation.uuid,
        engagement_address_type_uuid=base.employee_address_type.uuid,
        employee_address_type_uuid=base.employee_address_type.uuid,
    )
    address_classes_iterable = [x[1] for x in address_classes]

    ou_type_unit = gen_org_klass(
        facet_uuid=base.org_unit_type.uuid,
        title="Unit",
    )
    ou_level_l1 = gen_org_klass(
        facet_uuid=base.org_unit_level.uuid,
        title="L1",
    )

    man_type_man = gen_org_klass(
        facet_uuid=base.manager_type.uuid,
        title="Manager",
    )

    man_level_m1 = gen_org_klass(
        facet_uuid=base.manager_level.uuid,
        title="M1",
    )
    resp_r1 = gen_org_klass(
        facet_uuid=base.responsibility.uuid,
        title="R1",
    )

    not_primary = gen_org_klass(
        facet_uuid=base.primary_type.uuid,
        title="Non-primary",
    )

    visible = gen_org_klass(
        facet_uuid=base.visibility.uuid,
        title="Visible",
    )

    not_visible = gen_org_klass(
        facet_uuid=base.visibility.uuid,
        title="Not visible",
    )

    ea_assoc = gen_org_klass(
        facet_uuid=base.engagement_association_type.uuid,
        title="Association",
    )

    classes = [
        ou_type_unit,
        ou_level_l1,
        man_type_man,
        man_level_m1,
        resp_r1,
        not_primary,
        visible,
        not_visible,
        ea_assoc,
    ]
    classes.extend(address_classes_iterable)

    org_units, partial_managers = read_csv(
        path=base_path / "gir_orsted_hierarki_linje.csv",
        org_unit_type_uuid=ou_type_unit.uuid,
        org_unit_level_uuid=ou_level_l1.uuid,
    )

    org_units.extend(missing_ous(ou_type_unit.uuid, ou_level_l1.uuid))
    # department_no: uuid
    ou_mapping = {ou.user_key: ou.uuid for ou in flatten(org_units)}

    ou_fin = OrgUnit.from_simplified_fields(
        uuid=generate_uuid("w/e" + "fin"),
        user_key="financial_placeholder",
        name="Financial Organisation",
        org_unit_type_uuid=ou_type_unit.uuid,
        org_unit_level_uuid=ou_level_l1.uuid,
    )
    ou_legal = OrgUnit.from_simplified_fields(
        uuid=generate_uuid("w/e" + "legal"),
        user_key="legal_placeholder",
        name="Legal Organisation",
        org_unit_type_uuid=ou_type_unit.uuid,
        org_unit_level_uuid=ou_level_l1.uuid,
    )
    engagement_type_uuid_generator = UUIDGenerator("engagement_type")
    job_function_uuid_generator = UUIDGenerator("job_function")
    (
        employees,
        additional_engagement_ous,
        engagements,
        engagement_associations,
        addresses,
    ) = read_emd(
        path=base_path / "gir_from_emd.csv",
        org_unit_uuids=ou_mapping,
        primary_uuid=not_primary.uuid,
        engagement_type_uuid_generator=engagement_type_uuid_generator,
        job_function_uuid_generator=job_function_uuid_generator,
        visible_uuid=visible.uuid,
        not_visible_uuid=not_visible.uuid,
        address_klasses=address_classes,
        org_unit_type_uuid=ou_type_unit.uuid,
        financial_org_unit_uuid=ou_fin.uuid,
        legal_org_unit_uuid=ou_legal.uuid,
        engagement_association_type_uuid=ea_assoc.uuid,
        org_unit_level_uuid=ou_level_l1.uuid,
        org_uuid=base.organisation.uuid,
    )
    fixture2 = (
        *employees,
        *additional_engagement_ous,
        *engagements,
    )

    eng_mapping = {x.user_key: x for x in flatten(engagements)}
    managers = gen_managers(
        partial_managers=partial_managers,
        eng_mapping=eng_mapping,
        responsibility_uuid=resp_r1.uuid,
        manager_level_uuid=man_level_m1.uuid,
        manager_type_uuid=man_type_man.uuid,
    )
    # more shitty code:
    managers = list(map(lambda x: [x], managers))
    classes.extend(
        klasses_from_uuid_gen(
            facet_uuid=base.engagement_type.uuid,
            organisation_uuid=base.organisation.uuid,
            uuid_gen=engagement_type_uuid_generator,
        )
    )
    classes.extend(
        klasses_from_uuid_gen(
            facet_uuid=base.engagement_job_function.uuid,
            organisation_uuid=base.organisation.uuid,
            uuid_gen=job_function_uuid_generator,
        )
    )
    fixture = base_objs, classes, *org_units, [ou_fin], [ou_legal]
    # return (*fixture,)
    # return (*fixture2,)
    # return *engagement_associations, *managers, *addresses
    # return *managers[10:20], *addresses[10:20]
    # return (*managers,)
    # return (*addresses[9:],)
    return *fixture, *fixture2, *engagement_associations, *managers, *addresses
    # return org_units[:1]


async def main(base_path: Path):
    client = MoLoRaClient()
    values = read_values(base_path)
    async with client.context():
        for group in values:
            await client.load_objs(group)


if __name__ == "__main__":
    # os.environ["gir_base_path"]
    run(main(Path("/home/mw/gir")))



import click

from .fix_departments import unit_fixer
from .sd_changed_at import changed_at_cli
from .sd_changed_at_redo import cli as changed_at_redo
from .sd_fixup import cli as sd_fixup
from .sd_importer import cli as sd_importer
from .sync_job_id import sync_jobid
from .test_mo_against_sd import cli as mo_against_sd
from .test_sd_connectivity import check_connectivity


@click.group()
def SDTool():
    """Common entrypoint to SD programs."""
    pass


SDTool.add_command(sd_importer, "sd_importer")
SDTool.add_command(mo_against_sd, "test_mo_against_sd")
SDTool.add_command(check_connectivity)
SDTool.add_command(unit_fixer)
SDTool.add_command(sync_jobid)
SDTool.add_command(changed_at_cli)
SDTool.add_command(changed_at_redo)
SDTool.add_command(sd_fixup)


if __name__ == "__main__":
    SDTool()



import datetime
import logging
from abc import ABC
from abc import abstractmethod
from functools import lru_cache
from functools import partial
from operator import itemgetter
from typing import Union
from uuid import UUID

from more_itertools import ilen
from more_itertools import only
from more_itertools import pairwise
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.deprecation import deprecated
from ra_utils.load_settings import load_settings
from tqdm import tqdm


LOGGER_NAME = "updatePrimaryEngagements"
logger = logging.getLogger(LOGGER_NAME)


class MultipleFixedPrimaries(Exception):
    """Thrown when multiple fixed primaries are found doing recalculate.

    This means that a user entered invalid data in MO.
    """

    pass


class NoPrimaryFound(Exception):
    """Thrown when no primary is determined doing recalculate.

    This means the implementation specific backend did not fulfill the interface for
    the _find_primary method.
    """

    pass


def noop(*args, **kwargs):
    """Noop function, which consumes arguments and does nothing."""
    pass


class MOPrimaryEngagementUpdater(ABC):
    def __init__(self, settings=None, dry_run=False):
        self.settings = settings or load_settings()
        self.dry_run = dry_run

        self.helper = self._get_mora_helper(self.settings["mora.base"])

        # List of engagement filters to apply to check / recalculate respectively
        # NOTE: Should be overridden by subclasses
        self.check_filters = []
        self.calculate_filters = []

        self.primary_types, self.primary = self._find_primary_types()

    def _get_mora_helper(self, mora_base):
        """Construct a MoraHelper object.

        Split out solely to ease testing.
        """
        return MoraHelper(hostname=mora_base, use_cache=False)

    def _get_person(self, cpr=None, uuid=None, mo_person=None):
        """Fetch a person from MO.

        Only one of the given parameters should be given, if multiple are given
        UUID takes priority over CPR which in turn takes priority over mo_person.

        Args:
            cpr: The CPR number of the person.
            uuid: The MO uuid of the person.
            mo_person: An existing user object from MoraHelper.

        Returns:
            user object from MoraHelper.
        """

        @lru_cache(maxsize=None)
        def _get_org_uuid():
            org_uuid = self.helper.read_organisation()
            return org_uuid

        if uuid:
            mo_person = self.helper.read_user(user_uuid=uuid)
        elif cpr:
            mo_person = self.helper.read_user(
                user_cpr=cpr, org_uuid=self._get_org_uuid()
            )
        return mo_person

    def _read_engagement(self, user_uuid, date):
        """Fetch all engagements for user_uuid at date."""
        mo_engagements = self.helper.read_user_engagements(
            user=user_uuid,
            at=date,
            only_primary=True,  # Do not read extended info from MO.
            use_cache=False,
        )
        return mo_engagements

    @abstractmethod
    def _find_primary_types(self):
        """Find primary classes for the underlying implementation.

        Returns:
            2-tuple:
                primary_types: a dict from indirect primary names to UUIDs.
                    The used names are 'fixed_primary', 'primary' and 'non_primary',
                    as such these names should be keys in the dictionary.
                primary: a list of UUIDs that can considered to be primary.
                    Should be a subset of the values in primary_types.
        """
        raise NotImplementedError()

    @abstractmethod
    def _find_primary(self, mo_engagements):
        """Decide which of the engagements in mo_engagements is the primary one.

        This method does not need to handle fixed_primaries as the method will
        not be called if fixed_primaries exist.

        Args:
            mo_engagements: List of engagements

        Returns:
            UUID: The UUID of the primary engagement.
        """
        raise NotImplementedError()

    def _predicate_primary_is(self, primary_type_key, engagement):
        """Predicate on an engagements primary type.

        Example:

            is_non_primary = partial(_predicate_primary_is, "non_primary")
            mo_engagement = ...
            is_mo_engagement_non_primary = is_non_primary(mo_engagement)

        Args:
            primary_type_key: Lookup key into self.primary_types
            engagement: The engagement to check primary status from

        Returns:
            boolean: True if engagement has primary type equal to primary_type_key
                     False otherwise
        """
        assert primary_type_key in self.primary_types

        if not engagement.get("primary"):
            return False

        if engagement["primary"]["uuid"] == self.primary_types[primary_type_key]:
            logger.info(
                "Engagement {} is {}".format(engagement["uuid"], primary_type_key)
            )
            return True
        return False

    def _count_primary_engagements(self, check_filters, user_uuid, mo_engagements):
        """Count number of primaries.

        Args:
            check_filters: A list of predicate functions from (user_uuid, eng).
            user_uuid: UUID of the user to who owns the engagements.
            engagements: A list of MO engagements to count primaries from.

        Returns:
            3-tuple:
                engagement_count: Number of engagements processed.
                primary_count: Number of primaries found.
                filtered_primary_count: Number of primaries passing check_filters.
        """
        # Count number of engagements
        mo_engagements = list(mo_engagements)
        engagement_count = len(mo_engagements)

        # Count number of primary engagements, by filtering on self.primary
        primary_mo_engagements = list(
            filter(
                lambda eng: eng["primary"]["uuid"] in self.primary,
                mo_engagements,
            )
        )
        primary_count = len(primary_mo_engagements)

        # Count number of primary engagements, by filtering out special primaries
        # What consistutes a 'special primary' depend on the subclass implementation
        for filter_func in check_filters:
            primary_mo_engagements = filter(
                partial(filter_func, user_uuid), primary_mo_engagements
            )
        filtered_primary_count = ilen(primary_mo_engagements)

        return engagement_count, primary_count, filtered_primary_count

    def _check_user(self, check_filters, user_uuid):
        """Check the users primary engagement(s).

        Args:
            check_filters: A list of predicate functions from (user_uuid, eng).
            user_uuid: UUID of the user to check.

        Returns:
            Dictionary:
                key: Date at which the value is valid.
                value: A 3-tuple, from _count_primary_engagements.
        """
        # List of cut dates, excluding the very last one
        date_list = self.helper.find_cut_dates(uuid=user_uuid)
        date_list = date_list[:-1]
        # Map all our dates, to their corresponding engagements.
        mo_engagements = map(partial(self._read_engagement, user_uuid), date_list)
        # Map mo_engagements to primary counts
        primary_counts = map(
            partial(self._count_primary_engagements, check_filters, user_uuid),
            mo_engagements,
        )
        # Create dicts from cut_dates --> primary_counts
        return dict(zip(date_list, primary_counts))

    def _check_user_outputter(self, check_filters, user_uuid):
        """Check the users primary engagement(s).

        Args:
            check_filters: A list of predicate functions from (user_uuid, eng).
            user_uuid: UUID of the user to check.

        Returns:
            Generator of output 4-tuples:
                outputter: Function to output strings to
                string: The base output string
                user_uuid: User UUID for the output string
                date: Date for the output string
        """

        def to_output(e_count, p_count, fp_count):
            if e_count == 0:
                return (noop, "")
            if p_count == 0:
                return (print, "No primary")
            if p_count == 1:
                return (noop, "")
            if fp_count == 0:
                return (logger.info, "All primaries are special")
            if fp_count == 1:
                return (logger.info, "Only one non-special primary")
            return (print, "Too many primaries")

        user_results = self._check_user(check_filters, user_uuid)
        for date, (e_count, p_count, fp_count) in user_results.items():
            outputter, string = to_output(e_count, p_count, fp_count)
            yield outputter, string, user_uuid, date

    def _check_user_strings(self, check_filters, user_uuid):
        """Check the users primary engagement(s).

        Args:
            check_filters: A list of predicate functions from (user_uuid, eng).
            user_uuid: UUID of the user to check.

        Returns:
            Generator of output 2-tuples:
                outputter: Function to output strings to
                string: Formatted output string
        """
        outputs = self._check_user_outputter(check_filters, user_uuid)
        for outputter, string, user_uuid, date in outputs:
            final_string = string + " for {} at {}".format(user_uuid, date.date())
            yield outputter, final_string

    def check_user(self, user_uuid):
        """Check the users primary engagement(s).

        Prints messages to stdout / log as side-effect.

        Args:
            user_uuid: UUID of the user to check.

        Returns:
            None
        """
        outputs = self._check_user_strings(self.check_filters, user_uuid)
        for outputter, string in outputs:
            outputter(string)

    @deprecated
    def recalculate_primary(self, *args, **kwargs):
        """Deprecated alias for recalculate_user."""
        # Kept solely for backwards compatability
        return self.recalculate_user(*args, **kwargs)

    def _decide_primary(self, mo_engagements):
        """Decide which of the engagements in mo_engagements is the primary one.

        Args:
            mo_engagements: List of engagements

        Returns:
            2-tuple:
                UUID: The UUID of the primary engagement.
                primary_type_key: The type of the primary.
        """
        # First we attempt to find a fixed primary engagement.
        # If multiple are found, we throw an exception, as only one is allowed.
        # If one is found, it is our primary and we are done.
        # If none are found, we need to calculate the primary engagement.
        find_fixed_primary = partial(self._predicate_primary_is, "fixed_primary")

        # Iterator of UUIDs of engagements with primary = fixed_primary
        fixed_primary_engagement_uuids = map(
            itemgetter("uuid"), filter(find_fixed_primary, mo_engagements)
        )
        # UUID of engagement with primary = fixed_primary, exception or None
        fixed = only(
            fixed_primary_engagement_uuids, None, too_long=MultipleFixedPrimaries
        )
        if fixed:
            return fixed, "fixed_primary"

        # No fixed engagements, thus we must calculate the primary engagement.
        #
        # The calulcation of primary engagement depends on the underlying
        # implementation, thus we simply call self._find_primary here.
        primary = self._find_primary(mo_engagements)
        if primary:
            return primary, "primary"
        raise NoPrimaryFound()

    def _ensure_primary(self, engagement, primary_type_uuid, validity):
        """Ensure that engagement has the right primary_type.

        Assuming the engagement already has the correct primary_type this method
        is a noop, wheres if this is not the case an update request will be made
        against MO.

        Args:
            engagement: The engagement to (potentially) update.
            primary_type_uuid: The primary type to ensure the engagement has.
            validity: The validity of the change (if made).

        Returns:
            boolean: True if a change is made, False otherwise.
        """
        # Check if the required primary type is already set
        if engagement["primary"]["uuid"] == primary_type_uuid:
            logger.info(
                "No update as primary type is not changed: {}".format(validity["from"])
            )
            return False

        # At this point, we know that we have to update the engagement, thus we
        # construct an update payload and send it to MO.
        payload = {
            "type": "engagement",
            "uuid": engagement["uuid"],
            "data": {"primary": {"uuid": primary_type_uuid}, "validity": validity},
        }
        logger.debug("Edit payload: {}".format(payload))

        if not self.dry_run:
            response = self.helper._mo_post("details/edit", payload)
            assert response.status_code in (200, 400)
            if response.status_code == 400:
                # XXX: This shouldn't happen due to the previous check?
                logger.warn("Attempted edit, but no change needed.")
                return False
        return True

    def recalculate_user(self, user_uuid: Union[UUID, str], no_past=False):
        """(Re)calculate primary engagement for the entire history the user."""
        user_uuid = str(user_uuid)

        def fetch_mo_engagements(date):
            """Fetch engagements which are active at 'date' and fulfill our filters.

            Also ensures that the 'primary' attribute is set on all engagements.
            """

            def ensure_primary(engagement):
                """Ensure that engagement has a primary field."""
                # TODO: It would seem this happens for leaves, should we make a
                #       special type for this?
                # TODO: What does the above even mean? - Help?
                if not engagement["primary"]:
                    engagement["primary"] = {"uuid": self.primary_types["non_primary"]}
                return engagement

            # Fetch engagements
            mo_engagements = self._read_engagement(user_uuid, date)
            # Filter unwanted engagements
            for filter_func in self.calculate_filters:
                mo_engagements = filter(
                    partial(filter_func, user_uuid, no_past), mo_engagements
                )
            # Enrich engagements with primary, if required
            mo_engagements = map(ensure_primary, mo_engagements)
            mo_engagements = list(mo_engagements)

            return mo_engagements

        def calculate_validity(start, end):
            """Construct engagement primarity validity from start and end date."""
            to = datetime.datetime.strftime(
                end - datetime.timedelta(days=1), "%Y-%m-%d"
            )
            # Sentinel value for infinity is usually 9999-12-30 / 9999-12-31.
            # We assume anything above 9999-1-1 is sentinel value for infinity.
            if end >= datetime.datetime(9999, 1, 1, 0, 0):
                to = None
            validity = {
                "from": datetime.datetime.strftime(start, "%Y-%m-%d"),
                "to": to,
            }
            return validity

        logger.info("Calculate primary engagement: {}".format(user_uuid))
        number_of_edits = 0

        # Find a list of dates with changes in engagement, and for each change
        # decide which engagement is the primary between that and the next change.
        date_list = self.helper.find_cut_dates(user_uuid, no_past=no_past)
        for start, end in pairwise(date_list):
            logger.info("Recalculate primary, date: {}".format(start))

            mo_engagements = fetch_mo_engagements(start)
            logger.debug("MO engagements: {}".format(mo_engagements))

            # No engagements, no primary, and thus nothing to do
            if len(mo_engagements) == 0:
                continue

            # Decide which of the mo_engagements is the primary one, and also what
            # kind of primary it is, fixed_primary or just primary
            try:
                primary_uuid, primary_type_key = self._decide_primary(mo_engagements)
            except NoPrimaryFound:
                logger.warning(f"Unable to determine primary for {user_uuid}")
                primary_uuid = None

            validity = calculate_validity(start, end)

            # Update the primary type of all engagements (if required)
            for engagement in mo_engagements:
                # As there can only be one primary engagement at the time, all
                # engagements are non_primary by default.
                primary_type_uuid = self.primary_types["non_primary"]
                # Only the primary engagement is marked non_primary. The actual type
                # is simply the one provided by _decide_primary.
                if engagement["uuid"] == primary_uuid:
                    primary_type_uuid = self.primary_types[primary_type_key]

                changed = self._ensure_primary(engagement, primary_type_uuid, validity)
                if changed:
                    number_of_edits += 1

        return_dict = {user_uuid: number_of_edits}
        return return_dict

    def check_all(self):
        """Check all users for the existence of primary engagements."""
        print("Reading all users from MO...")
        all_users = self.helper.read_all_users()
        print("OK")
        for user in tqdm(all_users):
            self.check_user(user["uuid"])

    def recalculate_all(self, no_past=False):
        """Recalculate all users primary engagements."""
        print("Reading all users from MO...")
        all_users = self.helper.read_all_users()
        print("OK")
        edit_status = {}
        all_users = tqdm(all_users)
        all_users = map(itemgetter("uuid"), all_users)
        for user_uuid in all_users:
            try:
                status = self.recalculate_user(user_uuid, no_past=no_past)
                edit_status.update(status)
            except MultipleFixedPrimaries:
                print("{} has conflicting fixed primaries".format(user_uuid))
            except Exception as exp:
                print("Exception while processing {}: {}".format(user_uuid, exp))

        total_non_edits = 0
        total_edits = 0
        for number_of_edits in edit_status.values():
            if number_of_edits == 0:
                total_non_edits += 1
            total_edits += number_of_edits
        print("Total non-edits: {}".format(total_non_edits))
        print("Total edits: {}".format(total_edits))



#
# Copyright (c) Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
import logging

LOG_LEVEL = logging.DEBUG


def start_logging(log_file, detail_logging=None):
    if detail_logging is None:
        detail_logging = (
            "sdChangedAt",
            "sdCommon",
            "sdFixDepartments",
            "sdImport",
            "sdMox",
            "mora-helper",
        )

    for name in logging.root.manager.loggerDict:
        if name in detail_logging:
            logging.getLogger(name).setLevel(LOG_LEVEL)
        else:
            logging.getLogger(name).setLevel(logging.ERROR)

    logging.basicConfig(
        format="%(levelname)s %(asctime)s %(name)s %(message)s",
        level=LOG_LEVEL,
        filename=log_file,
    )



import logging
import re
from datetime import date
from datetime import datetime
from datetime import timedelta
from itertools import chain
from itertools import takewhile
from typing import Dict
from typing import Iterator
from typing import Optional
from typing import Tuple

from more_itertools import pairwise
from more_itertools import tabulate

from .sd_common import EmploymentStatus

# TODO: move constants elsewhere
# TODO: set back to "infinity" when MO can handle this
# MO_INFINITY: str = "infinity"

MO_INFINITY = None
SD_INFINITY: str = "9999-12-31"
DATE_REGEX_STR = "[0-9]{4}-(0[1-9]|1[0-2])-([0-2][0-9]|3[0-1])"

logger = logging.getLogger("sdImport")


def date_to_datetime(d: date) -> datetime:
    return datetime(d.year, d.month, d.day)


def format_date(date: datetime) -> str:
    return date.strftime("%Y-%m-%d")


def parse_date(date_str: str) -> datetime:
    return datetime.strptime(date_str, "%Y-%m-%d")


def _get_employment_from_date(
    employment: Dict, employment_date_as_engagement_start_date: bool
) -> datetime:
    # Make sure we do not have multiple EmploymentStatuses
    assert isinstance(employment["EmploymentStatus"], Dict)

    date = employment["EmploymentStatus"]["ActivationDate"]
    if employment_date_as_engagement_start_date:
        date = employment["EmploymentDate"]
    return parse_date(date)


def get_employment_dates(
    employment: Dict, employment_date_as_engagement_start_date: bool
) -> Tuple[datetime, datetime]:
    """
    Get the "from" and "to" date from the SD employment

    Args:
        employment: The SD employment
        employment_date_as_engagement_start_date: Use EmploymentDate as
            start engagement start date if True and use the activation date
            if False.

    Returns:
        Tuple containing the "from" and "to" dates.
    """

    status = EmploymentStatus(employment["EmploymentStatus"]["EmploymentStatusCode"])

    if status in EmploymentStatus.let_go():
        date_from = parse_date(employment["EmploymentDate"])
        termination_date = str(
            sd_to_mo_termination_date(employment["EmploymentStatus"]["ActivationDate"])
        )
        date_to = parse_date(termination_date)
    elif status == EmploymentStatus.Orlov:
        # We have seen examples where a leave begins BEFORE the SD
        # EmploymentDate which will cause the "assert" below to break
        # (see https://redmine.magenta-aps.dk/issues/48067#note-20)
        employment_date = _get_employment_from_date(employment, True)
        leave_activation_date = _get_employment_from_date(employment, False)
        date_from = min(employment_date, leave_activation_date)
        date_to = parse_date(employment["EmploymentStatus"]["DeactivationDate"])
    else:
        date_from = _get_employment_from_date(
            employment, employment_date_as_engagement_start_date
        )
        date_to = parse_date(employment["EmploymentStatus"]["DeactivationDate"])

    return date_from, date_to


# TODO: Create "MoValidity" and "SdValidity" classes based on the RA Models
#  "Validity" class and use these as input to the function below


def sd_to_mo_termination_date(sd_date: str) -> Optional[str]:
    """
    Convert SD termination date to MO termination date.

    In MO, the termination date is the last day of work, while in SD it is the
    first day of non-work.

    Args:
        sd_date: SD termination date formatted as "YYYY-MM-DD"

    Returns:
        MO termination date formatted as "YYYY-MM-DD"
    """

    assert isinstance(sd_date, str)
    assert re.compile(DATE_REGEX_STR).match(sd_date)

    if sd_date == SD_INFINITY:
        return MO_INFINITY

    # In MO, the termination date is the last day of work,
    # in SD it is the first day of non-work.
    _sd_date = parse_date(sd_date)
    mo_date = _sd_date - timedelta(days=1)

    return format_date(mo_date)


def to_midnight(dt: datetime) -> datetime:
    """Get previous midnight from datetime."""
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def is_midnight(dt: datetime) -> bool:
    """Check if datetime is at midnight."""
    return dt == to_midnight(dt)


def gen_cut_dates(from_datetime: datetime, to_datetime: datetime) -> Iterator[datetime]:
    """Get iterator of cut-dates between from_datetime and to_datetime.

    Args:
        from_datetime: the start date
        to_datetime: the end date

    Yields:
        The from_datetime, then all intermediate midnight datetimes and the to_datetime.
    """
    assert from_datetime < to_datetime

    # Tabulate to infinite iterator of midnights starting after from_datetime
    def midnight_at_offset(offset: int) -> datetime:
        return to_midnight(from_datetime) + timedelta(days=offset)

    midnights = takewhile(
        lambda midnight: midnight < to_datetime, tabulate(midnight_at_offset, start=1)
    )
    return chain([from_datetime], midnights, [to_datetime])


def gen_date_intervals(
    from_datetime: datetime, to_datetime: datetime
) -> Iterator[Tuple[datetime, datetime]]:
    """
    Get iterator capable of generating a sequence of datetime pairs
    incrementing one day at a time. The latter date in a pair is
    advanced by exactly one day compared to the former date in the pair.

    Args:
        from_datetime: the start date
        to_datetime: the end date

    Yields:
        The next date pair in the sequence of pairs
    """
    return pairwise(gen_cut_dates(from_datetime, to_datetime))



from copy import deepcopy
from unittest.mock import patch

from parameterized import parameterized

from sdlon.config import ChangedAtSettings
from sdlon.skip import cpr_env_filter
from .test_config import DEFAULT_CHANGED_AT_SETTINGS


class TestCprEnvFilter:
    @parameterized.expand(
        [
            (True, [], True),
            (True, ["5555555555"], False),
            (False, [], False),
            (False, ["5555555555"], True),
        ]
    )
    @patch("sdlon.skip.get_changed_at_settings")
    def test_return_true_for_exclude_mode_when_cpr_not_in_list(
        self, exclude_mode, cprs, expected, mock
    ):
        settings_dict = deepcopy(DEFAULT_CHANGED_AT_SETTINGS)
        settings_dict.update({"sd_exclude_cprs_mode": exclude_mode, "sd_cprs": cprs})
        mock.return_value = ChangedAtSettings.parse_obj(settings_dict)

        entity = {"PersonCivilRegistrationIdentifier": "5555555555"}

        assert cpr_env_filter(entity) == expected



import csv
import io
import logging
import re
import string
from functools import partial
from operator import itemgetter
from typing import List
from typing import Tuple

from more_itertools import flatten
from ra_utils.load_settings import load_setting
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.sql import column
from sqlalchemy.sql import table

from .ad_exceptions import ImproperlyConfigured
from .ad_reader import ADParameterReader
from .username_rules import method_2

logger = logging.getLogger(__name__)

NameType = List[str]


class UserNameGen:
    _setting_prefix = "integrations.ad_writer.user_names"

    @classmethod
    def get_implementation(cls) -> "UserNameGen":
        """Returns an implementation of `UserNameGen`.

        If you want to load occupied usernames from AD or elsewhere, you can
        call `load_occupied_names` on the instance returned by this method.

        By default, no occupied names are loaded on the instance returned by
        this method. This can be used in the cases where we want to spare the
        overhead of retrieving all AD usernames, for instance.
        """
        implementation_class = UserNameGenMethod2  # this is the default

        get_class_name = load_setting(f"{cls._setting_prefix}.class")

        try:
            name = get_class_name()
            implementation_class = cls._lookup_class_by_name(name)
        except (ValueError, FileNotFoundError):
            # ValueError: "not in settings file and no default"
            # FileNotFoundError: could not find "settings.json"
            logger.warning(
                "could not read settings, defaulting to %r",
                implementation_class,
            )
        except NameError:
            logger.warning(
                "could not find class %r, defaulting to %r",
                name,
                implementation_class,
            )

        instance = implementation_class()
        return instance

    @classmethod
    def _lookup_class_by_name(cls, name):
        if name and name in globals():
            return globals()[name]
        raise NameError(f"could not find class {name}")

    def __init__(self):
        self.occupied_names = set()
        self._loaded_occupied_name_sets = []

    def add_occupied_names(self, occupied_names: set) -> None:
        self.occupied_names.update(set(occupied_names))
        self._loaded_occupied_name_sets.append(occupied_names)

    def create_username(self, name: NameType, dry_run=False) -> str:
        raise NotImplementedError("must be implemented by subclass")

    def load_occupied_names(self):
        # Always load AD usernames when this method is called
        self.add_occupied_names(UserNameSetInAD())

        # Load any extra username sets specified in settings
        setting_name = f"{self._setting_prefix}.extra_occupied_name_classes"
        get_usernameset_class_names = load_setting(setting_name, default=[])
        try:
            usernameset_class_names = get_usernameset_class_names()
        except FileNotFoundError:  # could not find "settings.json"
            logger.info("could not read settings, not adding extra username sets")
        else:
            for name in usernameset_class_names:
                try:
                    cls = self._lookup_class_by_name(name)
                except NameError:
                    raise ImproperlyConfigured(
                        f"{setting_name!r} refers to unknown class {name!r}"
                    )
                else:
                    username_set = cls()
                    self.add_occupied_names(username_set)
                    logger.debug("added %r to set of occupied names", username_set)

    def is_username_occupied(self, username):
        return username.lower() in set(map(str.lower, self.occupied_names))


class UserNameGenMethod2(UserNameGen):
    """
    An implementation of metode 2 in the AD MOX specification document
    (Bilag: Tildeling af brugernavne).
    """

    def __init__(self):
        super().__init__()
        self.combinations = [
            method_2.FIRST,
            method_2.SECOND,
            method_2.THIRD,
            method_2.FOURTH,
            method_2.FIFTH,
            method_2.SIXTH,
        ]

    def _name_fixer(self, name: NameType) -> NameType:
        for i in range(0, len(name)):
            # Replace according to replacement list
            for char, replacement in method_2.CHAR_REPLACEMENT.items():
                name[i] = name[i].replace(char, replacement)

            # Remove all remaining characters outside a-z
            for char in name[i].lower():
                if ord(char) not in range(ord("a"), ord("z") + 1):
                    name[i] = name[i].replace(char, "")

        return name

    def _readable_combi(self, combi):
        readable_combi = []
        max_position = -1
        for i in range(0, len(combi)):
            # First name
            if combi[i] == "F":
                position = 0
            # First middle name
            if combi[i] == "1":
                position = 1
            # Second middle name
            if combi[i] == "2":
                position = 2
            # Third middle name
            if combi[i] == "3":
                position = 3
            # Last name (independent of middle names)
            if combi[i] == "L":
                position = -1
            if combi[i] == "X":
                position = None
            if position is not None and position > max_position:
                max_position = position
            readable_combi.append(position)
        return (readable_combi, max_position)

    def _create_from_combi(self, name, combi):
        """
        Create a username from a name and a combination.
        """
        (code, max_position) = self._readable_combi(combi)

        # Do not use codes that uses more names than the actual person has
        if max_position > len(name) - 2:
            return None

        # First letter is always first letter of first position
        if code[0] is not None:
            relevant_name = code[0]
            username = name[relevant_name][0].lower()
        else:
            username = "X"

        current_char = 0
        for i in range(1, len(code)):
            if code[i] == code[i - 1]:
                current_char += 1
            else:
                current_char = 0

            if code[i] is not None:
                relevant_name = code[i]
                if current_char >= len(name[relevant_name]):
                    username = None
                    break
                username += name[relevant_name][current_char].lower()
            else:
                username += "X"
        return username

    def create_username(self, name: NameType, dry_run: bool = False) -> str:
        """
        Create a new username in accodance with the rules specified in this file.
        The username will be the highest quality available and the value will be
        added to list of used names, so consequtive calles with the same name
        will keep returning new names until the algorithm runs out of options
        and a RuntimeException is raised.

        :param name: Name of the user given as a list with at least two elements.
        :param dry_run: If true the name will be returned, but will be added to
        the interal list of reserved names.
        :return: New username generated.
        """
        final_user_name = ""
        name = self._name_fixer(name)

        for permutation_counter in range(2, 10):
            for prioritation in range(0, len(self.combinations)):
                if final_user_name != "":
                    break
                i = 0
                for combi in self.combinations[prioritation]:
                    i = i + 1
                    username = self._create_from_combi(name, combi)
                    if not username:
                        continue
                    indexed_username = username.replace("X", str(permutation_counter))
                    if not self.is_username_occupied(indexed_username):
                        if not dry_run:
                            self.occupied_names.add(indexed_username)
                        final_user_name = indexed_username
                        return final_user_name

        # If we get to here, we completely failed to make a username
        raise RuntimeError("Failed to create user name")


class UserNameGenPermutation(UserNameGen):
    def __init__(self):
        super().__init__()
        self.length = 3
        self.consonants = "".join(set(string.ascii_lowercase) - set("aeiouy"))

    def create_username(self, name: NameType, dry_run: bool = False) -> str:
        suffix = 1
        while True:
            letters = self._extract_letters(name)
            new_username = "%s%d" % ("".join(letters), suffix)
            if not self.is_username_occupied(new_username):
                # An unused username was found, add it to the list of
                # occupied names and return.
                self.occupied_names.add(new_username)
                return new_username
            else:
                # We are still looking for an available username.
                # Bump the `suffix` variable.
                suffix += 1

    def _extract_letters(self, name: NameType):
        # Convert ["Firstname", "Last Name"] -> ["Firstname", "Last", "Name"]
        # and ["First-Name", "Last-Name"] -> ["First", "Name", "Last", "Name"]
        name = flatten(map(partial(re.split, r"[\-\s+]"), name))

        # Convert all name parts to lowercase
        name = list(map(str.lower, name))

        # Check name parts
        first_ascii = set(name[0]) & set(string.ascii_lowercase)
        next_consonant = set("".join(name[1:])) & set(self.consonants)
        assert len(name) > 0, "name must have at least one part"
        assert first_ascii, "first name part must contain at least one ASCII character"
        assert next_consonant, "next name parts must contain at least one consonant"

        def only(allowed: str, part: str):
            return "".join(ch for ch in part if ch.lower() in allowed)

        result = []

        # Take first letter of first name part (regardless of whether it is a vowel or
        # a consonant.)
        result.append(only(string.ascii_lowercase, name[0])[0])

        # Continue at first letter of the second name part (first part if only one part)
        p = min(1, len(name) - 1)  # second name part (or first if only one part)
        offset = 0  # = first letter

        while len(result) < self.length:
            part = name[p]
            try:
                result.append(only(self.consonants, part)[offset])
            except IndexError:
                # Check if there are still more name parts to use
                if p < len(name) - 1:
                    # If yes, use next name part, starting at first letter
                    p += 1
                    offset = 0
                else:
                    # If no, go back to first name
                    p = 0
            else:
                offset += 1

        return result


class UserNameSet:
    def __init__(self):
        self._usernames = set()

    def __contains__(self, username: str) -> bool:
        return username.lower() in set(map(str.lower, self._usernames))

    def __iter__(self):
        return iter(self._usernames)


class UserNameSetCSVFile(UserNameSet):
    _encoding = "utf-8-sig"
    _column_name = "Brugernavn"

    def __init__(self):
        get_path = load_setting("integrations.ad_writer.user_names.disallowed.csv_path")
        with io.open(get_path(), "r", encoding=self._encoding) as stream:
            reader = csv.DictReader(stream)
            self._usernames = set(row[self._column_name] for row in reader)


class UserNameSetInAD(UserNameSet):
    def __init__(self):
        reader = ADParameterReader()
        self._usernames = reader.get_all_samaccountname_values()


class UserNameSetInDatabase(UserNameSet):
    def __init__(self):
        connection_string, table_name, column_name = self._get_settings()
        session = self._get_session(connection_string)
        t = table(table_name, column(column_name))
        res = session.query(t).distinct()
        # Extract the usernames from returned list of tuples like [('Alice',), ('Bob',)]
        self._usernames = set(map(itemgetter(0), res))

    def _get_settings(self) -> Tuple[str, str, str]:
        get_table_name = load_setting(
            "integrations.ad_writer.user_names.disallowed.sql_table_name"
        )
        get_column_name = load_setting(
            "integrations.ad_writer.user_names.disallowed.sql_column_name"
        )
        get_connection_string = load_setting(
            "integrations.ad_writer.user_names.disallowed.sql_connection_string"
        )
        return get_connection_string(), get_table_name(), get_column_name()

    def _get_session(self, connection_string):
        engine = create_engine(connection_string)
        return sessionmaker(bind=engine, autoflush=False)()



import datetime
from datetime import date
from functools import partial
from operator import itemgetter
from typing import List
from typing import Optional
from typing import Tuple

import click
import httpx
from more_itertools import consume
from more_itertools import flatten
from more_itertools import one
from more_itertools import only
from more_itertools import side_effect
from more_itertools import unzip
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.apply import apply
from ra_utils.load_settings import load_setting
from tqdm import tqdm

from . import sd_payloads
from .config import get_importer_settings
from .sd_changed_at import ChangeAtSD
from .sd_common import EmploymentStatus
from .sd_common import mora_assert
from .sd_common import primary_types
from .sd_common import sd_lookup


def fetch_user_employments(cpr: str) -> List:
    # Notice, this will not get future engagements
    params = {
        "PersonCivilRegistrationIdentifier": cpr,
        "StatusActiveIndicator": "true",
        "StatusPassiveIndicator": "true",
        "DepartmentIndicator": "true",
        "EmploymentStatusIndicator": "true",
        "ProfessionIndicator": "true",
        "WorkingTimeIndicator": "true",
        "UUIDIndicator": "true",
        "SalaryAgreementIndicator": "false",
        "SalaryCodeGroupIndicator": "false",
        "EffectiveDate": date.today().strftime("%d.%m.%Y"),
    }

    sd_employments_response = sd_lookup("GetEmployment20111201", None, params)
    if "Person" not in sd_employments_response:
        return []

    employments = sd_employments_response["Person"]["Employment"]
    if not isinstance(employments, list):
        employments = [employments]

    return employments


def get_orgfunc_from_vilkaarligrel(
    class_uuid: str, mox_base: str = "http://localhost:8080"
) -> dict:
    url = f"{mox_base}/organisation/organisationfunktion"
    params = {"vilkaarligrel": class_uuid, "list": "true", "virkningfra": "-infinity"}
    r = httpx.get(url, params=params)
    r.raise_for_status()
    return only(r.json()["results"], default={})


def get_user_from_org_func(org_func: dict) -> Optional[str]:
    registrations = one(org_func["registreringer"])
    user = one(registrations["relationer"]["tilknyttedebrugere"])
    return user["uuid"]


def filter_missing_data(leave: dict) -> bool:
    return not one(leave["registreringer"])["relationer"].get("tilknyttedefunktioner")


def delete_orgfunc(uuid: str, mox_base: str = "http://localhost:8080") -> None:
    r = httpx.delete(f"{mox_base}/organisation/organisationfunktion/{uuid}")
    r.raise_for_status()


def fixup(ctx, mo_employees):
    def fetch_mo_engagements(mo_employee) -> dict:
        mo_uuid = mo_employee["uuid"]
        mo_engagements = mora_helper.read_user_engagement(user=mo_uuid, read_all=True)
        no_salary_mo_engagements = list(
            filter(
                lambda mo_engagement: mo_engagement["primary"]["user_key"] == "status0",
                mo_engagements,
            )
        )
        mo_salary_userkeys = map(itemgetter("user_key"), no_salary_mo_engagements)
        mo_dict = dict(zip(mo_salary_userkeys, no_salary_mo_engagements))
        return mo_dict

    def fetch_sd_employments(mo_employee):
        mo_cpr = mo_employee["cpr_no"]
        sd_employments = fetch_user_employments(mo_cpr)
        sd_ids = map(itemgetter("EmploymentIdentifier"), sd_employments)
        sd_dict = dict(zip(sd_ids, sd_employments))
        return sd_dict

    def fetch_pairs(mo_employee):
        try:
            mo_dict = fetch_mo_engagements(mo_employee)
            if not mo_dict:
                return None
            sd_dict = fetch_sd_employments(mo_employee)
            return mo_dict, sd_dict
        except Exception as exp:
            print(mo_employee)
            print(exp)
            return None

    def process_tuples(mo_dict, sd_dict):
        # Find intersection
        common_keys = mo_dict.keys() & sd_dict.keys()
        for key in common_keys:
            yield (key, mo_dict[key], sd_dict[key])

    def sd_not_status_0(work_tuple):
        key, mo_engagement, sd_employment = work_tuple
        sd_status = sd_employment["EmploymentStatus"]["EmploymentStatusCode"]
        return sd_status != "0"

    def generate_payload(work_tuple):
        key, mo_engagement, sd_employment = work_tuple
        print("Fixing", key)
        data = {
            "validity": mo_engagement["validity"],
            "primary": {"uuid": primary["non_primary"]},
        }
        payload = sd_payloads.engagement(data, mo_engagement)
        return payload

    mora_helper = ctx["mora_helper"]
    primary = primary_types(mora_helper)

    if ctx["progress"]:
        mo_employees = tqdm(mo_employees, unit="Employee")

    # Dict pair is an iterator of (dict, dict) tuples or None
    # First dict is a mapping from employment_id to mo_engagement
    # Second dict is a mapping from employment_id to sd_engagement
    dict_pairs = map(fetch_pairs, mo_employees)
    # Remove all the None's from dict_pairs
    dict_pairs = filter(None.__ne__, dict_pairs)

    # Convert dict_pairs into an iterator of three tuples:
    # (key, mo_engagement, sd_employment)
    # 'key' is the shared employment_id
    work_tuples = flatten(map(process_tuples, *unzip(dict_pairs)))
    # Filter all tuples, where the sd_employment has status 0
    work_tuples = filter(sd_not_status_0, work_tuples)
    # At this point, we have a tuple of items which need to be updated / fixed

    # Convert all the remaining tuples to MO payloads
    payloads = map(generate_payload, work_tuples)

    if ctx["dry_run"]:
        return

    for payload in payloads:
        response = mora_helper._mo_post("details/edit", payload)
        mora_assert(response)


@click.group()
@click.option("--mora-base", default="http://localhost:5000", help="URL for MO.")
@click.option("--json/--no-json", default=False, help="Output as JSON.")
@click.option("--progress/--no-progress", default=False, help="Print progress.")
@click.option("--fixup-status-0", default=False, help="Attempt to fix status-0 issues.")
@click.option(
    "--dry-run/--no-dry-run", default=False, help="Dry-run making no actual changes."
)
@click.pass_context
def cli(ctx, mora_base, **kwargs):
    """Tool to fixup MO entries according to SD data.

    This tool should never be needed, as it indicates issues in the main code.
    It is however needed due to the quality of the main code.
    """
    # ensure that ctx.obj exists and is a dict, no matter how it is called.
    ctx.ensure_object(dict)
    ctx.obj = dict(kwargs)
    ctx.obj["mora_helper"] = MoraHelper(hostname=mora_base, use_cache=False)


@cli.command()
@click.option(
    "--uuid", type=click.UUID, help="UUID of the user to check.", required=True
)
@click.pass_context
# Example UUID: 'fadfcc38-5d42-4857-a950-0adc65babb13'
def fixup_user(ctx, uuid):
    """Fix a single employee."""
    mo_employees = [ctx.obj["mora_helper"].read_user(user_uuid=uuid)]
    fixup(ctx.obj, mo_employees)


@cli.command()
@click.option(
    "--mox-base",
    default=load_setting("mox.base", "http://localhost:8080"),
    help="URL for Lora",
)
@click.pass_context
def fixup_leaves(ctx, mox_base):
    """Fix all leaves that are missing a link to an engagement."""
    settings = get_importer_settings()

    mora_helper = ctx.obj["mora_helper"]
    # Find all classes of leave_types
    leave_types, _ = mora_helper.read_classes_in_facet("leave_type")
    leave_type_uuids = map(itemgetter("uuid"), leave_types)

    # Get all leave objects
    orgfunc_getter = partial(get_orgfunc_from_vilkaarligrel, mox_base=mox_base)
    leave_objects = map(orgfunc_getter, leave_type_uuids)
    leave_objects = list(flatten(leave_objects))

    # Filter to get only those missing the 'engagement'.
    leave_objects = list(filter(filter_missing_data, leave_objects))
    leave_uuids = set(map(itemgetter("id"), leave_objects))
    # Delete old leave objects
    if ctx.obj["dry_run"]:
        click.echo(f"Dry-run. Would delete {len(leave_uuids)} leave objects")
    else:
        orgfunc_deleter = partial(delete_orgfunc, mox_base=mox_base)
        leave_uuids = tqdm(
            leave_uuids,
            unit="leave",
            desc="Deleting old leaves",
            disable=not ctx.obj["progress"],
        )
        consume(side_effect(orgfunc_deleter, leave_uuids))

    # Find all user uuids and cprs
    user_uuids = set(map(get_user_from_org_func, leave_objects))
    user_uuids = tqdm(
        user_uuids,
        unit="user",
        desc="Looking up users in MO",
        disable=not ctx.obj["progress"],
    )
    users = map(mora_helper.read_user, user_uuids)
    cpr_uuid_map = dict(map(itemgetter("cpr_no", "uuid"), users))
    # NOTE: This will only reimport current leaves, not historic ones
    #       This behavior is inline with sd_importer.py
    changed_at = ChangeAtSD(settings, datetime.datetime.now())

    def try_fetch_leave(cpr: str) -> Tuple[str, List[dict]]:
        """Attempt to lookup engagements from a CPR.

        Prints any errors but continues
        """
        employments = []
        try:
            employments = fetch_user_employments(cpr=cpr)
        except Exception as e:
            click.echo(e)
        # filter leaves
        leaves = list(
            filter(
                lambda employment: EmploymentStatus(
                    employment["EmploymentStatus"]["EmploymentStatusCode"]
                )
                == EmploymentStatus.Orlov,
                employments,
            )
        )
        return cpr, leaves

    cprs = tqdm(
        cpr_uuid_map.keys(),
        desc="Lookup users in SD",
        unit="User",
        disable=not ctx.obj["progress"],
    )
    leaves = dict(map(try_fetch_leave, cprs))

    # Filter users with leave
    leaves = dict(filter(apply(lambda cpr, engagement: engagement), leaves.items()))

    if ctx.obj["dry_run"]:
        click.echo(f"Dry-run. Would reimport leaves for {len(leaves)} users.")
        return

    for cpr, leaves in tqdm(
        leaves.items(),
        unit="user",
        desc="Reimporting leaves for users",
        disable=not ctx.obj["progress"],
    ):
        for leave in leaves:
            changed_at.create_leave(
                leave["EmploymentStatus"],
                leave["EmploymentIdentifier"],
                cpr_uuid_map[cpr],
            )


@cli.command()
@click.option(
    "--uuid", type=click.UUID, help="UUID of the organisation to check.", required=True
)
@click.pass_context
# Example UUID: 'ea5a237a-8f8b-4300-9a00-000006180002'
def fixup_department(ctx, uuid):
    """Fix all employees in an organisation."""
    mo_employees = (
        ctx.obj["mora_helper"].read_organisation_people(uuid, read_all=True).keys()
    )
    fixup(ctx.obj, mo_employees)


@cli.command()
@click.option(
    "--limit",
    type=click.INT,
    help="Number of employees to check. 0 --> All.",
    default=5,
)
@click.pass_context
def fixup_all(ctx, limit):
    """Fix limit/all employees in an MO instance."""
    mo_employees = ctx.obj["mora_helper"].read_all_users(limit=limit)
    fixup(ctx.obj, mo_employees)


if __name__ == "__main__":
    cli()



import re
from datetime import datetime
from operator import itemgetter
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import OrderedDict
from typing import Tuple

from .sd_common import ensure_list
from .sd_common import read_employment_at

INTERNAL_EMPLOYEE_REGEX = re.compile("[0-9]+")


def engagement_components(engagement_info) -> Tuple[str, Dict[str, List[Any]]]:
    employment_id = engagement_info["EmploymentIdentifier"]

    return employment_id, {
        "status_list": ensure_list(engagement_info.get("EmploymentStatus", [])),
        "professions": ensure_list(engagement_info.get("Profession", [])),
        "departments": ensure_list(engagement_info.get("EmploymentDepartment", [])),
        "working_time": ensure_list(engagement_info.get("WorkingTime", [])),
    }


def update_existing_engagement(
    sd_updater, mo_engagement, sd_engagement, person_uuid
) -> None:
    sd_updater.edit_engagement_department(sd_engagement, mo_engagement, person_uuid)
    sd_updater.edit_engagement_profession(sd_engagement, mo_engagement)
    sd_updater.edit_engagement_type(sd_engagement, mo_engagement)
    sd_updater.edit_engagement_worktime(sd_engagement, mo_engagement)


def create_engagement(sd_updater, employment_id, person_uuid) -> None:
    # Call SD to get SD employment
    sd_employment_payload = read_employment_at(
        datetime.now().date(), settings=sd_updater.settings, employment_id=employment_id
    )
    if sd_employment_payload is None:
        return

    assert not isinstance(sd_employment_payload, list)

    cpr = sd_employment_payload["PersonCivilRegistrationIdentifier"]
    sd_employment = sd_employment_payload.get("Employment")
    status = sd_employment.get("EmploymentStatus")  # type: ignore

    # Is it possible that sd_employment or status is None?...
    assert sd_employment
    assert status

    # Not sure what to do if several statuses are returned...
    assert not isinstance(status, list)

    # Call MO to create corresponding engagement in MO
    sd_updater.create_new_engagement(sd_employment, status, cpr, person_uuid)


def _is_external(employment_id: str) -> bool:
    """
    Check if the SD employee is an external employee. This is the
    case (at least in some municipalities...) if the EmploymentIdentifier
    contains letters.

    Args:
         employment_id: the SD EmploymentIdentifier

    Returns:
        True of the employment_id contains letters and False otherwise
    """

    match = INTERNAL_EMPLOYEE_REGEX.match(employment_id)
    return match is None


def is_employment_id_and_no_salary_minimum_consistent(
    engagement: OrderedDict, no_salary_minimum: Optional[int] = None
) -> bool:
    """
    Check that the external SD employees have JobPositionIdentifiers
    consistent with no_salary_limit
    (see https://os2web.atlassian.net/browse/MO-245).

    Args:
        engagement: the SD employment
        no_salary_minimum: the minimum allowed JobPositionIdentifier
          for external SD employees.

    Returns:
        True if the provided values are consistent and False otherwise.
    """

    if no_salary_minimum is None:
        return True

    employment_id, eng_components = engagement_components(engagement)
    professions = eng_components["professions"]
    if not professions:
        return True

    job_pos_ids_strings = map(itemgetter("JobPositionIdentifier"), professions)
    job_pos_ids = map(int, job_pos_ids_strings)

    def is_consistent(job_pos_id: int) -> bool:
        if _is_external(employment_id):
            return job_pos_id >= no_salary_minimum  # type: ignore
        return job_pos_id < no_salary_minimum  # type: ignore

    consistent = map(is_consistent, job_pos_ids)

    return all(consistent)



# Script for getting all changes in SD for a given person based on the persons
# employmentID or CPR number. The script extracts these info from the .tar.gz
# files in /opt/dipex/backup on the servers by parsing the files
# opt/dipex/os2mo-data-import-and-export/mo_integrations.log in the tar
# archives
import re
import sys
import tarfile
from enum import Enum
from enum import unique
from functools import partial
from pathlib import Path
from typing import List
from typing import Optional

import click
from lxml import etree
from lxml.etree import _Element  # as Element
from more_itertools import only
from pydantic import BaseModel

LOG_FILE = "opt/dipex/os2mo-data-import-and-export/mo_integrations.log"
BACKUP_DIR = "/opt/dipex/backup"
OUTPUT_FILE = "/tmp/analysis.txt"


@unique
class IdType(Enum):
    CPR = "PersonCivilRegistrationIdentifier"
    EMPLOYMENT_ID = "EmploymentIdentifier"


class SdPersonChange(BaseModel):
    # We use strings instead of datetimes since we only need the string
    # output anyway
    start_date: str
    end_date: str
    change: Optional[_Element] = None

    class Config:
        arbitrary_types_allowed = True


def get_tar_gz_archive_files(path: Path) -> List[Path]:
    """
    Get all .tar.gz files in the provided folder path.

    Args:
        path: The folder containing the tar.gz files

    Returns:
        List of tar.gz files in the provided folder
    """

    tars = list(filter(lambda f: f.is_file() and f.suffix == ".gz", path.iterdir()))
    tars.sort()
    return tars


def extract_log_file_lines(tar_gz_file: Path) -> List[str]:
    """
    Extract list of lines from the mo_integrations.log file.

    Args:
        tar_gz_file: The tar.gz file to extract the lines from.

    Returns:
        List of log files lines as UTF-8 strings.
    """
    with tarfile.open(str(tar_gz_file), "r:gz") as tar:
        try:
            log_file = tar.extractfile(LOG_FILE)
        except KeyError:
            return []

        if not log_file:
            return []

        # List of byte strings
        lines = log_file.readlines()
        uft8_lines = map(lambda line: line.decode("utf-8"), lines)
        lines_without_newline = map(lambda line: line.rstrip("\n"), uft8_lines)

        return list(lines_without_newline)


def get_sd_xml_responses(log_file_lines: List[str]) -> List[_Element]:
    """
    Get SD XML responses from list of log file lines.

    Args:
        log_file_lines: list of raw lines from log file.

    Returns:
         XML responses from SD
    """

    sd_response_lines = filter(lambda line: "sdCommon Response" in line, log_file_lines)

    # Only use the part of the string that is after
    # <?xml version="1.0" encoding="UTF-8" ?>
    xml_strings = map(lambda s: s.partition("?>")[2], sd_response_lines)

    # Convert for XML strings to XML element
    xml_element = map(lambda s: etree.fromstring(s), xml_strings)

    return list(xml_element)


def get_sd_person_changed_at_date_responses(
    sd_xml_responses: List[_Element],
) -> List[_Element]:
    return list(
        filter(lambda e: e.tag == "GetPersonChangedAtDate20111201", sd_xml_responses)
    )


def get_sd_person_changed(
    identifier_type: IdType, identifier: str, xml_root: _Element
) -> SdPersonChange:

    # Get the request date timestamps
    request_structure = xml_root.find("RequestStructure")
    activation_date = request_structure.find("ActivationDate").text.strip()
    deactivation_date = request_structure.find("DeactivationDate").text.strip()

    def id_match(id_type: IdType, id_: str, person: _Element) -> bool:
        if id_type == IdType.CPR:
            id_element = person.find(id_type.value)
        else:
            employment_element = person.find("Employment")
            id_element = employment_element.find(id_type.value)

        if id_ == id_element.text.strip():
            return True
        return False

    person_elements = xml_root.findall("Person")
    persons = filter(partial(id_match, identifier_type, identifier), person_elements)

    sd_person_changed = SdPersonChange(
        start_date=activation_date, end_date=deactivation_date, change=only(persons)
    )

    return sd_person_changed


def get_all_sd_person_changes(
    identifier_type: IdType, identifier: str, backup_folder: Path
) -> List[SdPersonChange]:
    tar_gz_files = get_tar_gz_archive_files(backup_folder)
    sd_person_changes = []
    for file in tar_gz_files:
        print(f"Analyzing {str(file.name)}...")
        log_file_lines = extract_log_file_lines(file)
        sd_xml_responses = get_sd_xml_responses(log_file_lines)
        changed_at_date_responses = get_sd_person_changed_at_date_responses(
            sd_xml_responses
        )
        sd_person_changes += [
            get_sd_person_changed(identifier_type, identifier, sd_response)
            for sd_response in changed_at_date_responses
        ]
    return sd_person_changes


def output_to_file(all_changes: List[SdPersonChange], file: Path) -> None:
    with open(file, "w") as fp:
        non_empty_changes = filter(
            lambda person: person.change is not None, all_changes
        )
        for sd_person_change in non_empty_changes:
            fp.write(40 * "-" + "\n")
            fp.write(f"Start date: {sd_person_change.start_date}\n")
            fp.write(f"End date: {sd_person_change.end_date}\n\n")
            fp.write(
                etree.tostring(sd_person_change.change, pretty_print=True).decode(
                    "utf-8"
                )
            )


@click.group()
def cli():
    pass


@cli.command()
@click.option(
    "--folder",
    type=click.Path(exists=True, readable=True),
    default=Path(BACKUP_DIR),
    help="Folder containing the .tar.gz backup files",
)
@click.option("--cpr", type=click.STRING, help="The CPR number of the SD person")
@click.option(
    "--eid", type=click.STRING, help="The SD employment identifier of the person"
)
@click.option(
    "--output",
    type=click.Path(writable=True, file_okay=True),
    default=Path(OUTPUT_FILE),
    help="The output file containing the changes for the given person",
)
def analyze(folder, cpr, eid, output):
    if not ((cpr is None) ^ (eid is None)):
        sys.exit("You must provide exactly one of CPR or employmentID")

    if cpr:
        cpr_regex = re.compile("[0-9]{10}")
        if not cpr_regex.match(cpr):
            sys.exit("CPR must be exactly 10 digits (no hyphen)")
        all_changes = get_all_sd_person_changes(IdType.CPR, cpr, Path(folder))
    else:
        eid_regex = re.compile("[0-9]{5}")
        if not eid_regex.match(eid):
            sys.exit("CPR must be exactly 5 digits (no hyphen)")
        all_changes = get_all_sd_person_changes(IdType.EMPLOYMENT_ID, eid, Path(folder))
    output_to_file(all_changes, Path(output))


if __name__ == "__main__":
    cli()



import datetime

from integrations.calculate_primary.common import logger
from integrations.calculate_primary.common import MOPrimaryEngagementUpdater
from integrations.SD_Lon.sdlon import sd_common


class SDPrimaryEngagementUpdater(MOPrimaryEngagementUpdater):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def remove_past(user_uuid, no_past, eng):
            if no_past and eng["validity"]["to"]:
                to = datetime.datetime.strptime(eng["validity"]["to"], "%Y-%m-%d")
                if to < datetime.datetime.now():
                    return False
            return True

        def remove_no_salary(eng):
            return not self._predicate_primary_is("no_salary", eng)

        def remove_no_salary_check(user_uuid, eng):
            return remove_no_salary(eng)

        def remove_no_salary_calculate(user_uuid, no_past, eng):
            return remove_no_salary(eng)

        def remove_missing_user_key(user_uuid, no_past, eng):
            return "user_key" in eng

        self.check_filters = [
            remove_no_salary_check,
        ]

        self.calculate_filters = [
            remove_past,
            remove_no_salary_calculate,
            remove_missing_user_key,
        ]

    def _find_primary_types(self):
        # Keys are; fixed_primary, primary, no_salary and non-primary
        primary_types = sd_common.primary_types(self.helper)
        primary = [
            primary_types["fixed_primary"],
            primary_types["primary"],
            primary_types["no_salary"],
        ]
        return primary_types, primary

    def _find_primary(self, mo_engagements):
        def non_integer_userkey(mo_engagement):
            try:
                # non-integer user keys should universally be status0, and as such
                # they should already have been filtered out, thus if they have not
                # been filtered out, they must have the wrong primary_type.
                int(mo_engagement["user_key"])
            except ValueError:
                self._fixup_status_0(mo_engagement)
                # Filter it out, as it should have been
                return False
            return True

        # Ensure that all mo_engagements have integer user_keys.
        mo_engagements = list(filter(non_integer_userkey, mo_engagements))

        if not mo_engagements:
            return None

        # The primary engagement is the engagement with the highest occupation rate.
        # - The occupation rate is found as 'fraction' on the engagement.
        #
        # If two engagements have the same occupation rate, the tie is broken by
        # picking the one with the lowest user-key integer.
        primary_engagement = max(
            mo_engagements,
            # Sort first by fraction, then reversely by user_key integer
            key=lambda eng: (eng.get("fraction") or 0, -int(eng["user_key"])),
        )
        return primary_engagement["uuid"]

    def _fixup_status_0(self, mo_engagement):
        logger.warning("Engagement type not status0. Will fix.")
        validity = mo_engagement["validity"]
        payload = {
            "type": "engagement",
            "uuid": mo_engagement["uuid"],
            "data": {
                "primary": {"uuid": self.primary_types["no_salary"]},
                "validity": validity,
            },
        }
        logger.debug("Status0 edit payload: {}".format(payload))
        if not self.dry_run:
            response = self.helper._mo_post("details/edit", payload)
            assert response.status_code == 200
            logger.info("Status0 fixed")



import uuid
from datetime import datetime
from unittest import mock

import los_files
import los_stam
import uuids
from hypothesis import given
from mox_helpers.mox_helper import ElementNotFound
from pydantic import Field

from .helpers import HelperMixin
from .helpers import mock_config
from .helpers import mock_create_mox_helper
from .strategies import csv_buf_from_model


class MockStamCSV(los_stam.StamCSV):
    mock_field: str = Field(default="testing testing")

    @staticmethod
    def get_filename() -> str:
        return "mock-filename.csv"

    @staticmethod
    def get_facet_bvn() -> str:
        return "mock_facet_bvn"

    @property
    def bvn(self) -> str:
        return "bvn"

    @property
    def title(self) -> str:
        return "title"

    @property
    def class_uuid(self) -> uuid.UUID:
        return uuid.uuid4()


class TestStamImporterLoadCSVIfNewer(HelperMixin):
    """Test `los_stam.StamImporter._load_csv_if_newer`."""

    _datetime_last_imported = datetime(2020, 1, 1)
    _datetime_last_modified = datetime(2020, 1, 2)

    def test_returns_csv_if_file_is_newer(self):
        # Pretend that `get_modified_datetime_for_file` finds a file, and that
        # `read_csv` returns its parsed contents.
        with self._mock_get_modified_datetime(
            return_value=self._datetime_last_modified
        ):
            with self._mock_read_csv(MockStamCSV()):
                result = self._run_load_csv_if_newer()
                assert len(result) == 1
                assert isinstance(result[0], MockStamCSV)
                assert result[0].mock_field == MockStamCSV().mock_field

    def test_returns_nothing_if_file_is_older(self):
        # Pretend that `get_modified_datetime_for_file` finds a file, but its
        # modified date is equal to the date of the last import run. In that
        # case, return nothing.
        with self._mock_get_modified_datetime(
            return_value=self._datetime_last_imported
        ):
            result = self._run_load_csv_if_newer()
            assert result is None

    def test_handles_missing_file(self):
        # Pretend that no file is found by `get_modified_datetime_for_file`.
        with self._mock_get_modified_datetime(side_effect=ValueError()):
            result = self._run_load_csv_if_newer()
            assert result is None

    def test_create_classes_from_csv(self):
        # Pretend we are adding exactly one LoRa class based on a single-item CSV file
        rows = [MockStamCSV()]

        with mock_config():
            instance = los_stam.StamImporter(self._datetime_last_imported)
            with mock_create_mox_helper(los_stam) as mh:
                mox_helper = mh.return_value
                mox_helper.read_element_klassifikation_facet.return_value = uuid.uuid4()
                self._run_until_complete(
                    instance._create_classes_from_csv(MockStamCSV, rows)
                )
                # Assert that we add one class in the facet
                mox_helper.insert_klassifikation_klasse.assert_called_once()

                # Assert that we convert the facet UUID to string, and we use the right
                # facet UUID.
                (
                    payload,
                    class_uuid,
                ) = mox_helper.insert_klassifikation_klasse.call_args.args
                expected_facet_uuid = str(
                    mox_helper.read_element_klassifikation_facet.return_value
                )
                actual_facet_uuid = payload["relationer"]["facet"][0]["uuid"]
                assert expected_facet_uuid == actual_facet_uuid

                # Assert that we convert the class UUID to string
                assert isinstance(class_uuid, str)

    def test_get_or_create_facet_existing_facet(self):
        with mock_config():
            with mock_create_mox_helper(los_stam) as mh:
                expected_facet_uuid = uuid.uuid4()
                mox_helper = mh.return_value

                # Pretend that we find the facet UUID in LoRa (= it already exists)
                mox_helper.read_element_klassifikation_facet.return_value = (
                    expected_facet_uuid
                )

                result = self._run_until_complete(
                    los_stam.StamImporter._get_or_create_facet(MockStamCSV, mox_helper)
                )

                # Assert that we retrieve the facet UUID in LoRa by looking up the BVN
                mox_helper.read_element_klassifikation_facet.assert_called_once_with(
                    bvn=MockStamCSV.get_facet_bvn()
                )
                # Assert that we got the facet UUID that we mocked
                assert result == expected_facet_uuid

    def test_get_or_create_facet_new_facet(self):
        with mock_config():
            with mock_create_mox_helper(los_stam) as mh:
                mox_helper = mh.return_value
                expected_facet_uuid = uuids.uuid_gen(MockStamCSV.get_facet_bvn())

                # Pretend that looking up the facet by its BVN will fail
                mox_helper.read_element_klassifikation_facet.side_effect = (
                    ElementNotFound
                )

                result = self._run_until_complete(
                    los_stam.StamImporter._get_or_create_facet(MockStamCSV, mox_helper)
                )

                # Assert that we look for the facet UUID in LoRa by looking up the BVN
                mox_helper.read_element_klassifikation_facet.assert_called_once_with(
                    bvn=MockStamCSV.get_facet_bvn()
                )
                # Assert that we added a new facet to LoRa
                mox_helper.insert_klassifikation_facet.assert_called_once_with(
                    mock.ANY,
                    expected_facet_uuid,
                )
                # Assert that we got the facet UUID that we constructed ourselves
                assert result == expected_facet_uuid

    def _run_load_csv_if_newer(self):
        instance = los_stam.StamImporter(self._datetime_last_imported)
        return instance._load_csv_if_newer(MockStamCSV)


class TestParseEngagementstypeCSV:
    @given(csv_buf_from_model(model=los_stam.Engagementstype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Engagementstype)


class TestParseEnhedstypeCSV:
    @given(csv_buf_from_model(model=los_stam.Enhedstype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Enhedstype)


class TestParseITSystemCSV:
    @given(csv_buf_from_model(model=los_stam.ITSystem))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.ITSystem)


class TestParseStillingsbetegnelseCSV:
    @given(csv_buf_from_model(model=los_stam.Stillingsbetegnelse))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Stillingsbetegnelse)


class TestParseBVNStillingsbetegnelseCSV:
    @given(csv_buf_from_model(model=los_stam.BVNStillingsbetegnelse))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.BVNStillingsbetegnelse)


class TestParseLederansvarCSV:
    @given(csv_buf_from_model(model=los_stam.Lederansvar))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Lederansvar)


class TestParseLederniveauCSV:
    @given(csv_buf_from_model(model=los_stam.Lederniveau))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Lederniveau)


class TestParseLedertypeCSV:
    @given(csv_buf_from_model(model=los_stam.Ledertype))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Ledertype)


class TestParseTilknytningsrolleCSV:
    @given(csv_buf_from_model(model=los_stam.Tilknytningsrolle))
    def test_parse_raises_nothing(self, csv_buf):
        los_files.parse_csv(csv_buf.readlines(), los_stam.Tilknytningsrolle)



import asyncio
from datetime import date
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

import config
import tqdm
from aiohttp import ClientSession
from aiohttp import ClientTimeout
from aiohttp import TCPConnector
from more_itertools import chunked
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import MoxHelper
from os2mo_helpers.mora_helpers import MoraHelper
from ra_utils.headers import TokenSettings


def get_tcp_connector():
    settings = config.get_config()
    return TCPConnector(limit=settings.max_concurrent_requests)


def get_client_session():
    return ClientSession(
        connector=get_tcp_connector(), timeout=ClientTimeout(total=None)
    )


async def create_details(
    session: ClientSession, detail_payloads: Iterable[dict]
) -> None:
    """Helper function for submitting create detail payloads"""
    url = "/service/details/create"
    await submit_payloads(session, url, detail_payloads, "create details")


async def edit_details(session: ClientSession, detail_payloads: Iterable[dict]) -> None:
    """Helper function for submitting edit detail payloads"""
    url = "/service/details/edit"
    await submit_payloads(session, url, detail_payloads, "edit details")


async def terminate_details(
    session: ClientSession,
    detail_payloads: Iterable[dict],
    ignored_http_statuses: Optional[Tuple[int]] = (404,),
) -> None:
    """Helper function for submitting terminate detail payloads"""
    url = "/service/details/terminate"
    await submit_payloads(
        session,
        url,
        detail_payloads,
        "terminate details",
        ignored_http_statuses=ignored_http_statuses,
    )


async def create_it(payload: dict, obj_uuid: str, mox_helper: MoxHelper) -> None:
    """Helper function for reating an IT system"""
    await mox_helper.insert_organisation_itsystem(payload, obj_uuid)


async def create_klasse(payload: dict, obj_uuid: str, mox_helper: MoxHelper) -> None:
    """Helper function for creating a Klasse object"""
    await mox_helper.insert_klassifikation_klasse(payload, obj_uuid)


async def submit_payloads(
    session: ClientSession,
    endpoint: str,
    payloads: Iterable[dict],
    description: str,
    ignored_http_statuses: Optional[Tuple[int]] = None,
) -> None:
    """
    Send a list of payloads to OS2mo. The payloads are chunked based on preset variable
    and submitted concurrently.

    :param session: A aiohttp session
    :param endpoint: Which endpoint to send the payloads to
    :param payloads: An iterable of dict payloads
    :param description: A description to print as part of the output
    """
    settings = config.get_config()
    base_url = settings.mora_base
    headers = TokenSettings().get_headers()

    async def submit(data: List[dict]) -> None:
        # Use semaphore to throttle the amount of concurrent requests
        async with session.post(
            base_url + endpoint,
            params={"force": 1},
            json=list(data),
            headers=headers,
        ) as response:
            if ignored_http_statuses and response.status in ignored_http_statuses:
                print(f"{endpoint} returned status {response.status}, ignoring")
            else:
                response.raise_for_status()

    chunks = chunked(payloads, settings.os2mo_chunk_size)
    tasks = list(map(submit, chunks))
    if len(tasks) == 0:
        return

    for f in tqdm.tqdm(
        asyncio.as_completed(tasks), total=len(tasks), unit="chunk", desc=description
    ):
        await f


async def lookup_organisationfunktion():
    """Helper function for fetching all available 'organisationfunktion' objects."""
    settings = config.get_config()
    mox = await create_mox_helper(settings.mox_base)
    return await mox.search_organisation_organisationfunktion(params={"bvn": "%"})


def lookup_employees():
    """Helper function for fetching all available 'employee' objects."""
    settings = config.get_config()
    mh = MoraHelper(hostname=settings.mora_base, export_ansi=True)
    return mh.read_all_users()


def convert_validities(from_time: date, to_time: date) -> Tuple[str, Optional[str]]:
    from_time_str = from_time.isoformat()
    to_time_str = to_time.isoformat()
    return from_time_str, to_time_str if to_time_str != "9999-12-31" else None



import time
import uuid
from datetime import datetime
from functools import partial
from itertools import chain
from random import choice
from random import randint
from unittest import TestCase

import requests

from ..ad_sync import AdMoSync
from ..ad_writer import ADWriter
from ..read_ad_conf_settings import read_settings
from ..user_names import UserNameGen
from ..utils import AttrDict
from ..utils import recursive_dict_update
from .name_simulator import create_name


class MOTestMixin(object):
    """Mixin to connect to MO and check service configuration.

    Used by the MOTestCase class.
    """

    def get_mo_host(self):
        """Get the base URI for the MO test instance.

        Examples:
            http://localhost:5000
            https://moratest.magenta.dk

        Returns:
            str: URI including schema for the MO test instance.
        """
        return "http://localhost:5000"

    def _fetch_mo_service_configuration(self):
        """Fetch the /service/configuration endpoint on MO.

        Example:
            {
                "read_only": false,
                ...,
                "show_user_key": true
            }

        Returns:
            dict: The JSON response from MO as a dict or None
        """
        host = self.get_mo_host()
        url = host + "service/configuration"
        response = requests.get(url)
        if response.status_code != 200:
            return None
        return response.json()

    def _check_mo_ready_for_testing(self):
        """Check if a MO instance can be reached and is in readonly mode.

        Example:
            "Unable to reach MO instance"

        Returns:
            str: The reason why MO is not reading for testing or None if ready.
        """
        configuration = self._fetch_mo_service_configuration()
        read_only_key = "read_only"
        if configuration is None:
            return "Unable to reach MO instance"
        elif read_only_key not in configuration:
            return "MO instance did not return readonly status"
        if configuration[read_only_key] is False:
            # Consider putting MO into read-only mode using:
            # curl -X PUT -H 'Content-Type: application/json' \
            #      -d '{"status": true}' http://localhost:5000/read_only/
            return "MO instance is NOT readonly"
        return None


class MOTestCase(TestCase, MOTestMixin):
    """TestCase, which verifies MO connection in setUP()."""

    def setUp(self):
        # Fetch MO status, and skipTest if any issues are found.
        status = self._check_mo_ready_for_testing()
        if status:
            self.skipTest(status)


def dict_modifier(updates):
    """Wrapper around recursive_dict_update, which provides updates beforehand.

    Example:

        updates = {'action': 'set'}

        function_with_expects_single_argument_transformer(
            dict_modifier(updates)
        )

    Args:
        updates: dictionary with updates to be applied later, when the returned
                 function is actually called.

    Returns:
        function: A partially applied recursive_dict_update function, waiting
                  for the original to apply updates to.
    """
    return partial(recursive_dict_update, updates=updates)


def mo_modifier(updates):
    """Wrapper around recursive_dict_update, which provides updates beforehand.

    Example:

        updates = {'action': 'set'}

        function_with_expects_mo_values_call(
            mo_modifier(updates)
        )

    Args:
        updates: dictionary with updates to be applied later, when the returned
                 function is actually called.

    Returns:
        function: A partially applied recursive_dict_update function, waiting
                  for the original to apply updates to.
    """

    def mo_mod(mo_values, *args, **kwargs):
        return recursive_dict_update(mo_values, updates=updates)

    return mo_mod


class ADWriterTestSubclass(ADWriter):
    """Testing subclass of ADWriter."""

    def __init__(
        self,
        read_ad_information_from_mo,
        ad_values_func=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        # List of scripts to be executed via run_ps
        self.scripts = []
        # Transformer for mo_values return
        self.read_ad_information_from_mo = read_ad_information_from_mo
        # Replace real `_find_ad_user` with mock
        if kwargs.get("mock_find_ad_user", True):
            self._find_ad_user = lambda ad_user, ad_dump=None: ad_values_func()

    def _init_name_creator(self):
        """Mocked to pretend no names are occupied.

        This method would normally use ADReader to read usernames from AD.
        """
        self.name_creator = UserNameGen.get_implementation()

    def _create_session(self):
        """Mocked to return a fake-class which writes scripts to self.scripts.

        This method would normally send scripts to powershell via WinRM.
        """

        def run_ps(ps_script):
            # Add our script to the list
            self.scripts.append(ps_script)
            # Fake the WinRM run_ps return type
            return AttrDict(
                {
                    "status_code": 0,
                    "std_out": b"",
                    "std_err": b"",
                }
            )

        # Fake the WinRM session object
        return AttrDict(
            {
                "run_ps": run_ps,
            }
        )

    def _get_retry_exceptions(self):
        """Mocked to return an empty list, i.e. never retry.

        This method would normally return the WinRM transport exception, to
        cause retrying to happen.
        """
        return []

    def read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        raise NotImplementedError("Should be overridden in __init__")

    def _read_ad_information_from_mo(self, uuid, read_manager=True, ad_dump=None):
        return super().read_ad_information_from_mo(uuid, read_manager, ad_dump)


def _no_transformation(default, *args, **kwargs):
    return default


class TestADMixin(object):
    # Useful for local testing
    generate_dynamic_person = True
    default_person = None

    def _prepare_dynamic_person(self):
        def random_date():
            unixtime = randint(1, int(time.time()))
            return datetime.fromtimestamp(unixtime).strftime("%d%m%y")

        def random_digit():
            return choice("0123456789")

        def random_digits(num_digits):
            return "".join(random_digit() for _ in range(num_digits))

        def gen_cpr():
            return random_date() + random_digits(4)

        dynamic_person = {
            "uuid": str(uuid.uuid4()),
            "name": create_name(),
            "nickname": create_name(),
            "cpr": gen_cpr(),
            "employment_number": random_digits(4),
            "read_manager": True,
            "manager_name": create_name(),
            "manager_cpr": gen_cpr(),
        }
        return dynamic_person

    def _prepare_static_person(self):
        static_person = {
            "uuid": "7ccbd9aa-gd60-4fa1-4571-0e6f41f6ebc0",
            "name": ("Martin", "Lee", "Gore"),
            "nickname": ("Depeche", "Mode"),
            "employment_number": "101",
            "cpr": "1122334455",
            "read_manager": True,
            "manager_name": ("Daniel", "Miller"),
            "manager_cpr": "1122334455",
        }
        return static_person

    def _prepare_person(self, person_transformer=None, *args, **kwargs):
        if self.default_person is None:
            default_person = self._prepare_static_person()
            if self.generate_dynamic_person:
                default_person = self._prepare_dynamic_person()
            # Add computed fields
            creator = UserNameGen.get_implementation()
            sam_account_name = creator.create_username(list(default_person["name"]))
            default_person.update(
                **{
                    "full_name": " ".join(default_person["name"]),
                    "sam_account_name": sam_account_name,
                    "manager_sam": default_person["manager_name"][0],
                    "manager_email": default_person["manager_name"][0] + "@magenta.dk",
                }
            )
            # Add static fields
            default_person.update(
                **{
                    "end_date": "2089-11-11",
                    "title": "Musiker",
                    "unit": "Enhed",
                    "unit_uuid": "101bd9aa-0101-0101-0101-0e6f41f6ebc0",
                    "unit_user_key": "Musik",
                    "unit_postal_code": "8210",
                    "unit_city": "Aarhus N",
                    "unit_streetname": "Fahrenheit 451",
                    "location": "Kommune\\Forvalting\\Enhed\\",
                    "level2orgunit": "Ingen",
                    "forvaltning": "Beskæftigelse, Økonomi & Personale",
                }
            )
            transformer_func = person_transformer or _no_transformation
            self.default_person = transformer_func(default_person, *args, **kwargs)
        return self.default_person

    def _prepare_mo_values(self, mo_values_transformer=None, *args, **kwargs):
        person = self._prepare_person()
        # Convert raw person data into mo_values data
        person["name"] = [" ".join(person["name"][:-1]), person["name"][-1]]
        person["nickname"] = [
            " ".join(person["nickname"][:-1]),
            person["nickname"][-1],
        ]
        person["manager_name"] = " ".join(person["manager_name"])

        #        if not read_manager:
        #            del person['manager_name']
        #            del person['manager_sam']
        #            del person['manager_email']
        #            del person['manager_cpr']
        #           person["read_manager"] = False

        transformer_func = mo_values_transformer or _no_transformation
        return transformer_func(person, *args, **kwargs)

    def _prepare_get_from_ad(self, ad_transformer, *args, **kwargs):
        person = self._prepare_person()
        # Convert raw person data into ad_values data
        default_ad_person = {
            "ObjectGUID": person["uuid"],
            "SID": {
                "AccountDomainSid": {
                    "AccountDomainSid": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx",
                    "BinaryLength": 24,
                    "Value": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx",
                },
                "BinaryLength": 28,
                "Value": "S-x-x-xx-xxxxxxxxxx-xxxxxxxxxx-xxxxxxxxxx-xxxxx",
            },
            "PropertyCount": 11,
            "PropertyNames": [
                "ObjectGUID",
                "SID",
                "DistinguishedName",
                "Enabled",
                "GivenName",
                "Name",
                "ObjectClass",
                "SamAccountName",
                "Surname",
                "UserPrincipalName",
                "extensionAttribute1",
            ],
            "DistinguishedName": "CN="
            + person["full_name"]
            + ",OU="
            + person["unit"]
            + ",DC=lee",
            "Enabled": True,
            "Name": person["full_name"],
            "ObjectClass": "user",
            "SamAccountName": person["sam_account_name"],
            "GivenName": person["name"][-1],
            "UserPrincipalName": "_".join(person["full_name"]).lower() + "@magenta.dk",
            "extensionAttribute1": person["cpr"],
            "AddedProperties": [],
            "ModifiedProperties": [],
            "RemovedProperties": [],
        }
        transformer_func = ad_transformer or _no_transformation
        return transformer_func(default_ad_person, *args, **kwargs)

    def _prepare_settings(self, early_settings_transformer=None):
        """Load default settings for AD tests.

        Args:
            early_settings_transformer: Function to transform settings.

        Returns:
            dict: Default settings after transformation.
        """
        default_settings = {
            "integrations.ad": [
                {
                    "cpr_field": "cpr_field",
                    "cpr_separator": "ad_cpr_sep",
                    "system_user": "system_user",
                    "password": "password",
                    "properties": [],
                    "search_base": "search_base",
                    "integrations.ad.ad_mo_sync_mapping": {},
                    "ad_mo_sync_terminate_missing": False,
                    "ad_mo_sync_terminate_missing_require_itsystem": True,
                    "ad_mo_sync_terminate_disabled": True,
                    "ad_mo_sync_pre_filters": [],
                    "ad_mo_sync_terminate_disabled_filters": [],
                    "servers": ["server123"],
                }
            ],
            "integrations.ad.winrm_host": "dummy",
            # "integrations.ad.sam_filter": "sam_filter",
            "mora.base": "http://example.org",
            "integrations.ad.write.uuid_field": "uuid_field",
            "integrations.ad.write.level2orgunit_field": "level2orgunit_field",
            "integrations.ad.write.org_unit_field": "org_field",
            "integrations.ad.write.upn_end": "epn_end",
            "integrations.ad.write.level2orgunit_type": "level2orgunit_type",
            "integrations.ad_writer.template_to_ad_fields": {
                "Name": "{{ mo_values['full_name'] }} - {{ user_sam }}",
                "Displayname": "{{ mo_values['name'][0] }} {{ mo_values['name'][1] }}",
                "GivenName": "{{ mo_values['name'][0] }}",
                "SurName": "{{ mo_values['name'][1] }}",
                "EmployeeNumber": "{{ mo_values['employment_number'] }}",
            },
            "address.visibility.public": "address_visibility_public_uuid",
            "address.visibility.internal": "address_visibility_internal_uuid",
            "address.visibility.secret": "address_visibility_secret_uuid",
        }
        transformer_func = early_settings_transformer or _no_transformation
        modified_settings = transformer_func(default_settings)
        for ad_settings in modified_settings["integrations.ad"]:
            ad_settings["properties"] = list(
                map(
                    lambda x: x.lower(),
                    chain(
                        modified_settings.get(
                            "integrations.ad_writer.template_to_ad_fields", {}
                        ).keys(),
                        modified_settings.get(
                            "integrations.ad_writer.mo_to_ad_fields", {}
                        ).values(),
                        [modified_settings["integrations.ad.write.org_unit_field"]],
                        [
                            modified_settings[
                                "integrations.ad.write.level2orgunit_field"
                            ]
                        ],
                        [modified_settings["integrations.ad.write.uuid_field"]],
                    ),
                )
            )
        return modified_settings


class TestADWriterMixin(TestADMixin):
    def _setup_adwriter(
        self,
        late_transform_settings=None,
        transform_mo_values=None,
        early_transform_settings=None,
        transform_ad_values=lambda x: x,
        **kwargs,
    ):
        transformer_func = late_transform_settings or _no_transformation
        self.settings = transformer_func(
            read_settings(self._prepare_settings(early_transform_settings))
        )
        self.mo_values_func = partial(self._prepare_mo_values, transform_mo_values)
        self.ad_values_func = partial(self._prepare_get_from_ad, transform_ad_values)
        self.ad_writer = ADWriterTestSubclass(
            all_settings=self.settings,
            read_ad_information_from_mo=self.mo_values_func,
            ad_values_func=self.ad_values_func,
            **kwargs,
        )


class AdMoSyncTestSubclass(AdMoSync):
    def __init__(
        self,
        mo_values_func,
        mo_e_username_func,
        ad_values_func,
        mo_seed_func,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.mo_values = mo_values_func()
        self.ad_values = ad_values_func()
        self.e_username = mo_e_username_func()

        self.mo_seed = mo_seed_func()
        self.mo_post_calls = []

    def _verify_it_systems(self):
        pass

    def _setup_mora_helper(self):
        def get_e_details(detail_type):
            return lambda *args, **kwargs: self.mo_seed[detail_type]

        def _mo_post(url, payload, force=True):
            # Register the call, so we can test against it
            self.mo_post_calls.append({"url": url, "payload": payload, "force": force})
            # response.text --> "OK"
            return AttrDict({"text": "OK", "raise_for_status": lambda: None})

        def read_user(uuid):
            return self.mo_values

        def update_user(uuid, data):
            payload = {"type": "employee", "uuid": uuid, "data": data}
            return _mo_post("details/edit", payload)

        return AttrDict(
            {
                "read_organisation": lambda: "org_uuid",
                "read_classes_in_facet": lambda x: [
                    [
                        {"uuid": "address_visibility_public_uuid"},
                        {"uuid": "address_visibility_internal_uuid"},
                        {"uuid": "address_visibility_secret_uuid"},
                    ]
                ],
                "read_user": read_user,
                "read_all_users": lambda: [self.mo_values],
                "read_user_engagement": get_e_details("engagement"),
                "get_e_addresses": get_e_details("address"),
                "get_e_itsystems": get_e_details("it"),
                "update_user": update_user,
                "_mo_post": _mo_post,
            }
        )

    def _setup_ad_reader_and_cache_all(self, index, cache_all=True):
        def read_user(cpr, cache_only):
            # We only support one person in our mocking
            if cpr != self.mo_values["cpr"]:
                raise NotImplementedError("Outside mocking")
            # If we got that one person, return it
            return self.ad_values

        def get_settings():
            return self.settings["integrations.ad"][0]

        ad_reader = AttrDict(
            {
                "_get_setting": get_settings,
                "read_user": read_user,
            }
        )
        return ad_reader


class TestADMoSyncMixin(TestADMixin):
    def _initialize_configuration(self):
        def ident(x, *args, **kwargs):
            return x

        self.settings = self._prepare_settings(ident)
        self.mo_values_func = partial(self._prepare_mo_values, ident)
        self.ad_values_func = partial(self._prepare_get_from_ad, ident)
        self.mo_e_username_func = lambda: ""
        self.mo_seed_func = lambda: {}

    def _setup_admosync(
        self,
        transform_settings=None,
        transform_mo_values=None,
        transform_ad_values=None,
        seed_e_username=None,
        seed_mo=None,
    ):
        if transform_settings:
            self.settings = self._prepare_settings(transform_settings)
        if transform_mo_values:
            self.mo_values_func = partial(self._prepare_mo_values, transform_mo_values)
        if transform_ad_values:
            self.ad_values_func = partial(
                self._prepare_get_from_ad, transform_ad_values
            )
        if seed_e_username:
            self.mo_e_username_func = seed_e_username
        if seed_mo:
            self.mo_seed_func = seed_mo

        self.ad_sync = AdMoSyncTestSubclass(
            all_settings=self.settings,
            mo_values_func=self.mo_values_func,
            ad_values_func=self.ad_values_func,
            mo_e_username_func=self.mo_e_username_func,
            mo_seed_func=self.mo_seed_func,
        )



from uuid import uuid4

import los_pers
from hypothesis import given
from hypothesis import strategies as st

from .helpers import mock_config


class TestPersonImporter:
    _azid_it_system_uuid = uuid4()

    @given(st.builds(los_pers.Person))
    def test_generate_employee_az_id_payload_uses_setting(self, person):
        importer = self._get_configured_importer()
        payload = importer.generate_employee_az_id_payload(person, "2021-11-17")
        assert payload["itsystem"]["uuid"] == str(self._azid_it_system_uuid)

    def _get_configured_importer(self):
        with mock_config(azid_it_system_uuid=self._azid_it_system_uuid):
            return los_pers.PersonImporter()



import functools
import json
import logging
import os
import pathlib
from abc import ABC, abstractmethod
from collections import defaultdict
from enum import Enum
import datetime

import requests
from sqlalchemy.orm import sessionmaker

from exporters.sql_export.lc_for_jobs_db import get_engine
from exporters.sql_export.sql_table_defs import KLE

from ra_utils.load_settings import load_settings
from os2mo_data_import.helpers import MoraHelper
LOG_FILE = 'opgavefordeler.log'

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO, filename=LOG_FILE)


class Aspects(Enum):
    Indsigt = 1
    Udfoerende = 2
    Ansvarlig = 3


# Maps between the enum and scopes on the classes from the aspect facet
ASPECT_MAP = {
    "INDSIGT": Aspects.Indsigt,
    "UDFOERENDE": Aspects.Udfoerende,
    "ANSVARLIG": Aspects.Ansvarlig,
}


def get_session(engine):
    return sessionmaker(bind=engine, autoflush=False)()


class KLEAnnotationIntegration(ABC):
    """Import and export of KLE annotation from or to an external source."""

    # XXX: This uses a simple inheritance based pattern. We might want to use
    # something like a Strategy here. However, maybe YAGNI.

    def __init__(self):
        
        self.settings = load_settings()

        self.mora_base = self.settings.get("mora.base")
        self.mora_session = self._get_mora_session(token=os.environ.get("SAML_TOKEN"))
        self.helper = MoraHelper(hostname=self.settings.get('mora.base'))
        self.org_uuid = self.helper.read_organisation()

        kle_classes = self.get_kle_classes_from_mo()
        self.kle_uuid_map = {item["user_key"]: item["uuid"] for item in kle_classes}

        aspect_classes = self.get_aspect_classes_from_mo()
        self.aspect_map = {
            ASPECT_MAP[clazz["scope"]]: clazz["uuid"] for clazz in aspect_classes
        }

    def _get_mora_session(self, token) -> requests.Session:
        s = requests.Session()
        if token is not None:
            s.headers.update({"SESSION": token})
        s.headers.update({"SESSION": token})
        s.verify = False
        return s

    def _get_mo_org_uuid(self) -> str:
        """
        Get the UUID of the organisation configured in OS2mo
        :return:
        """
        logger.info("Fetching Organisation UUID from OS2mo")
        r = self.mora_session.get("{}/service/o/".format(self.mora_base))
        r.raise_for_status()
        return r.json()[0]["uuid"]

    def get_kle_classes_from_mo(self) -> list:
        """Get all of the kle_number 'klasse' objects from OS2mo"""
        logger.info("Fetching KLE numbers from OS2mo")
        kle_numbers,  _ = self.helper.read_classes_in_facet('kle_number')
        return kle_numbers

    def get_aspect_classes_from_mo(self) -> list:
        """Get all of the kle_aspect 'klasse' objects from OS2mo"""
        logger.info("Fetching KLE aspect classes from OS2mo")
        kle_aspects,  _ = self.helper.read_classes_in_facet('kle_aspect')
        return kle_aspects

    def get_all_org_units_from_mo(self) -> list:
        """Get a list of all units from OS2mo"""
        logger.info("Fetching all org units from OS2mo")
        url = "{}/service/o/{}/ou/".format(self.mora_base, self.org_uuid)
        r = self.mora_session.get(url)
        r.raise_for_status()
        units = r.json()["items"]

        logger.info("Found {} units".format(len(units)))
        return units

    def post_payloads_to_mo(self, payloads: list):
        """Submit a list of details payloads to OS2mo"""
        logger.info("Posting payloads to OS2mo ")
        url = "{}/service/details/create".format(self.mora_base)

        r = self.mora_session.post(url, json=payloads, params={"force": 1})
        r.raise_for_status()

    @abstractmethod
    def run(self):
        """Implement this, normally to execute import or export."""
        pass


class OpgavefordelerImporter(KLEAnnotationIntegration):
    def __init__(self):
        super().__init__()
        self.opgavefordeler_url = self.settings.get(
            "integrations.os2opgavefordeler.url"
        )
        self.opgavefordeler_session = self._get_opgavefordeler_session(
            token=self.settings.get("integrations.os2opgavefordeler.token")
        )
        self.lc_session = get_session(get_engine())

    def _get_opgavefordeler_session(self, token) -> requests.Session:
        s = requests.Session()
        s.headers.update({"Authorization": "Basic {}".format(token)})
        return s

    def get_kle_info_from_opgavefordeler(self, kle_numbers: list) -> list:
        """Get all KLE-number info from OS2opgavefordeler

        This will give information on which unit is 'Ansvarlig' for a certain
        KLE-number.
        The API will perform inheritance and deduce the unit logically responsible
        for a certain number if no unit is directly responsible,
        so the result is filtered of all duplicates

        The resulting data has one entry per KLE-number and is on the form:
        [
          {
            "kle": {
              "number": "81",
              "name": "Kommunens personale"
            },
            "org": {
              "manager": {
                "name": "...",
                "email": "...",
                "esdhId": "c162842f-8036-411c-95ac-9c81042e9530",
                "esdhName": null,
                "initials": "...",
                "jobTitle": "..."
              },
              "businessKey": "99e2521a-c52d-4dfa-838f-b205184f3c00",
              "name": "...",
              "esdhId": "99e2521a-c52d-4dfa-838f-b205184f3c00",
              "esdhName": "...",
              "email": null,
              "phone": null,
              "foreignKeys": {}
            },
            "employee": null
          },
        ]
        """
        logger.info("Fetching KLE info from OS2opgavefordeler")
        url = "{}/TopicRouter/api".format(self.opgavefordeler_url)
        s = self.opgavefordeler_session

        unit_data = []
        for key in kle_numbers:
            try:
                r = s.get(url, params={"kle": key})
                r.raise_for_status()
                unit_data.append(r.json())
            except requests.exceptions.HTTPError:
                logger.warning("KLE number '{}' not found".format(key))

        seen_keys = set()
        filtered = []
        for item in unit_data:
            key = item["kle"]["number"]
            if key not in seen_keys:
                filtered.append(item)
                seen_keys.add(key)

        logger.info("Found {} items".format(len(filtered)))
        return filtered

    def get_org_unit_info_from_opgavefordeler(self, org_units_uuids: list) -> list:
        """Get all org-unit info from OS2opgavefordeler

        This will give information about which KLE-numbers the unit has a
        'Udførende' and 'Indsigt' relationship with.

        Empty results are filtered

        The resulting data is on the form:
        [
          [
            "5d6fe93d-2cf0-4bc7-b29a-68c14f9ae681",
            {
              "PERFORMING": [
                "15"
              ],
              "INTEREST": [
                "15"
              ]
            }
          ],
          ...
        ]
        """
        logger.info("Fetching org unit info from OS2opgavefordeler")
        url = "{}/TopicRouter/api/ou/{}"
        s = self.opgavefordeler_session
        org_unit_info = {}
        for uuid in org_units_uuids:
            try:
                r = s.get(url.format(self.opgavefordeler_url, uuid))
                r.raise_for_status()
                org_unit_info[uuid] = r.json()
                logger.debug("Adding {}".format(uuid))
            except requests.exceptions.HTTPError:
                continue

        def filter_empty(item):
            info = item[1]
            return info["INTEREST"] or info["PERFORMING"]

        filtered = list(filter(filter_empty, org_unit_info.items()))

        logger.info("Found {} items".format(len(filtered)))
        return filtered

    def get_ansvarlig_tuples(self):
        """Fetch KLE info from opgavefordeler and generate list of tuples
        mapping org unit uuids to KLE number uuids for the 'ansvarlig' aspect"""

        opgavefordeler_ansvarlig = self.get_kle_info_from_opgavefordeler(
            list(self.kle_uuid_map.keys())
        )
        return [
            (item["org"]["businessKey"], self.kle_uuid_map.get(item["kle"]["number"]))
            for item in opgavefordeler_ansvarlig
        ]

    def get_indsigt_and_udfoerende_tuples(self):
        """Fetch Org unit info from opgavefordeler and generate lists of tuples
        mapping org unit uuids to KLE number uuids for the 'indsigt' and
        'udfoerende' aspects"""

        org_unit_uuids = [unit["uuid"] for unit in self.get_all_org_units_from_mo()]
        org_unit_info = self.get_org_unit_info_from_opgavefordeler(org_unit_uuids)

        indsigt = [
            (org_unit_uuid, self.kle_uuid_map.get(kle))
            for org_unit_uuid, info in org_unit_info
            for kle in info.get("INTEREST")
            if kle in self.kle_uuid_map
        ]

        udfoerende = [
            (org_unit_uuid, self.kle_uuid_map.get(kle))
            for org_unit_uuid, info in org_unit_info
            for kle in info.get("PERFORMING")
            if kle in self.kle_uuid_map
        ]

        return indsigt, udfoerende

    def build_org_unit_maps(self):
        """Build a map of opgavefordeler data

        For each of Ansvarlig, Indsigt and Udførende, generate a map between
        org units and KLE-numbers they are associated to
        """

        def tuple_list_to_dict(d: dict, t: tuple):
            org_uuid, kle_uuid = t
            d.setdefault(org_uuid, []).append(kle_uuid)
            return d

        ansvarlig = self.get_ansvarlig_tuples()
        indsigt, udfoerende = self.get_indsigt_and_udfoerende_tuples()

        ansvarlig, indsigt, udfoerende = map(
            lambda x: functools.reduce(tuple_list_to_dict, x, dict()),
            [ansvarlig, indsigt, udfoerende],
        )

        return ansvarlig, indsigt, udfoerende

    def build_delete_payload(self, mo_uuid):
        return {
            "uuid": mo_uuid,
            "validity": {"from": "1920-01-01", "to": str(datetime.date.today())},
            "type": "kle",
        }

    def build_create_payload(self, org_unit_uuid, kle_number, kle_aspects):
        kle_payload = self.build_kle_payload(org_unit_uuid, kle_number, kle_aspects)

        payload = {
            **kle_payload,
            "type": "kle",
        }
        return payload

    def build_edit_payload(self, org_unit_uuid, kle_number, kle_aspects, obj_uuid):
        kle_payload = self.build_kle_payload(org_unit_uuid, kle_number, kle_aspects)

        payload = {
            "data": {
                **kle_payload,
            },
            "uuid": obj_uuid,
            "type": "kle",
        }
        return payload

    def build_kle_payload(self, org_unit_uuid, kle_uuid, kle_aspects):

        aspects_uuids = [self.aspect_map[aspect] for aspect in kle_aspects]

        payload = {
            "org_unit": {"uuid": org_unit_uuid},
            "kle_aspect": [{"uuid": uuid} for uuid in aspects_uuids],
            "kle_number": {"uuid": kle_uuid},
            "validity": {"from": "1920-01-01", "to": None},
        }
        return payload

    def convert_mo_kle_to_org_unit_map(self, mo_kle):
        """Convert KLE objects from SQL cache to a similar org_unit_map
        format as Opgavefordeler data, with added OS2mo object UUIDs

        The general structure is listed below

        {
            <ORG_UNIT_UUID>: {
                <KLE_NUMBER_UUID>: (
                    <MO_OBJECT_UUID>,
                    {
                        Aspects.XXX,
                        Aspects.YYY,
                    }
                ),
                ...
            },
            ...
        }
        """

        org_unit_map = {}

        aspect_map = {
            'Udførende': Aspects.Udfoerende,
            'Indsigt': Aspects.Indsigt,
            'Ansvarlig': Aspects.Ansvarlig,
        }

        for row in mo_kle:
            org_unit = org_unit_map.setdefault(row.enhed_uuid, {})
            _, aspects = org_unit.setdefault(row.kle_nummer_uuid, (row.uuid, set()))
            aspect = aspect_map[row.kle_aspekt_titel]
            aspects.add(aspect)

        return org_unit_map

    def convert_opgavefordeler_to_org_unit_map(
        self, ansvarlig: dict, indsigt: dict, udfoerende: dict
    ):
        """Convert the three aspect mappings to a map between org unit uuids,
        kle uuids and the associated aspects
        {
            <ORG_UNIT_UUID>: {
                <KLE_NUMBER_UUID>: {
                    Aspects.XXX,
                    Aspects.YYY,
                }
                ...
            },
            ...
        }
        """
        aspect_maps = {
            Aspects.Ansvarlig: ansvarlig,
            Aspects.Indsigt: indsigt,
            Aspects.Udfoerende: udfoerende,
        }

        org_unit_map = {}
        for aspect, aspect_map in aspect_maps.items():
            for org_uuid, kle_uuids in aspect_map.items():
                for kle_uuid in kle_uuids:
                    org_unit_map.setdefault(org_uuid, {}).setdefault(
                        kle_uuid, set()
                    ).add(aspect)

        return org_unit_map

    def create_diff(self, ansvarlig, indsigt, udfoerende):
        """Compare Opgavefordeler data with existing OS2mo data and
        determine what should be deleted, created and updated"""

        mo_kle = self.lc_session.query(KLE).all()
        mo_kle_org_unit_map = self.convert_mo_kle_to_org_unit_map(mo_kle)

        org_unit_map = self.convert_opgavefordeler_to_org_unit_map(
            ansvarlig, indsigt, udfoerende
        )

        # Compare every existing OS2mo KLE object with opgavefordeler
        # to see if an entry exists for every org_unit/kle_number pair.
        # If no longer in map, mark object UUID for deletion
        deleted_uuids = {
            row.uuid
            for row in mo_kle
            if not org_unit_map.get(row.enhed_uuid, {}).get(row.kle_nummer_uuid)
        }
        deleted = [self.build_delete_payload(uuid) for uuid in deleted_uuids]

        new = []
        updated = []

        # Compare every org_unit/kle_number pair in opgavefordeler map, with
        # OS2mo data.
        # If no data exists in OS2mo, mark as new
        # If data exists, but is different set of aspects, mark as update
        # else skip
        for unit, kle_info in org_unit_map.items():
            for kle_number, aspects in kle_info.items():
                mo_info_tuple = mo_kle_org_unit_map.get(unit, {}).get(kle_number)
                if not mo_info_tuple:
                    # New object
                    new.append(self.build_create_payload(unit, kle_number, aspects))
                else:
                    mo_obj_uuid, mo_aspects = mo_info_tuple
                    if mo_aspects != aspects:
                        # Updated object
                        updated.append(
                            self.build_edit_payload(
                                unit, kle_number, aspects, mo_obj_uuid
                            )
                        )

        return deleted, new, updated

    def handle_new(self, create_payloads: list):
        """Create KLE org functions"""

        logger.info("{} new KLE objects".format(len(create_payloads)))
        url = "{}/service/details/create".format(self.mora_base)

        r = self.mora_session.post(url, json=create_payloads, params={"force": 1})
        r.raise_for_status()

    def handle_update(self, edit_payloads: list):
        """Edit existing KLE org functions"""
        logger.info("{} updated KLE objects".format(len(edit_payloads)))
        url = "{}/service/details/edit".format(self.mora_base)

        r = self.mora_session.post(url, json=edit_payloads, params={"force": 1})
        r.raise_for_status()

    def handle_delete(self, delete_payloads: list):
        """Terminate KLE org functions"""

        logger.info("{} KLE objects to be terminated".format(len(delete_payloads)))
        url = "{}/service/details/terminate".format(self.mora_base)

        r = self.mora_session.post(url, json=delete_payloads, params={"force": 1})
        r.raise_for_status()

    def run(self):
        logger.info("Starting import")

        # Generate a map of data from opgavefordeler
        ansvarlig, indsigt, udfoerende = self.build_org_unit_maps()

        # Generate a diff towards OS2mo
        deleted, new, updated = self.create_diff(ansvarlig, indsigt, udfoerende)

        self.handle_new(new)
        self.handle_update(updated)
        self.handle_delete(deleted)

        logger.info("Done")


if __name__ == "__main__":
    importer = OpgavefordelerImporter()
    importer.run()



import os
import csv
import sys
import datetime
from pathlib import Path
from chardet.universaldetector import UniversalDetector

from os2mo_data_import import ImportHelper
import viborg_uuids
sys.path.append('..')
import sd_importer
import ad_reader

MUNICIPALTY_NAME = os.environ.get('MUNICIPALITY_NAME', 'SD-Løn Import')
MUNICIPALTY_CODE = os.environ.get('MUNICIPALITY_CODE', 0)
MOX_BASE = os.environ.get('MOX_BASE', 'http://localhost:8080')
MORA_BASE = os.environ.get('MORA_BASE', 'http://localhost:80')
MANAGER_FILE = os.environ.get('MANAGER_FILE', 'Organisationsdata.csv')

# ORIGIN FOR TESTS WIH ACTUAL API
# GLOBAL_GET_DATE = datetime.datetime(2006, 1, 1, 0, 0) # will not work
# GLOBAL_GET_DATE = datetime.datetime(2009, 1, 1, 0, 0)
# GLOBAL_GET_DATE = datetime.datetime(2011, 1, 1, 0, 0)
# GLOBAL_GET_DATE = datetime.datetime(2014, 2, 15, 0, 0)
GLOBAL_GET_DATE = datetime.datetime(2019, 2, 15, 0, 0)
# GLOBAL_GET_DATE = datetime.datetime(2019, 5, 10, 0, 0)

importer = ImportHelper(
    create_defaults=True,
    mox_base=MOX_BASE,
    mora_base=MORA_BASE,
    system_name='SD-Import',
    end_marker='SDSTOP',
    store_integration_data=True,
    seperate_names=True
)

ad_usernames, ad_uuids = viborg_uuids.read_ad_and_uuids()

detector = UniversalDetector()
with open(MANAGER_FILE, 'rb') as csvfile:
    for row in csvfile:
        detector.feed(row)
        if detector.done:
            break
detector.close()
encoding = detector.result['encoding']

manager_rows = []
with open(MANAGER_FILE, encoding=encoding) as csvfile:
    reader = csv.DictReader(csvfile, delimiter=';')
    for row in reader:
        if row['Leder 1 (cpr-nummer)']:
            new_row = {
                'cpr': row['Leder 1 (cpr-nummer)'].replace('-', ''),
                'ansvar' : row['Lederansvar "Leder 1"'],
                'afdeling': row['SD kort navn (afd.kode)']
            }
            manager_rows.append(new_row)

        if row['Leder 2 (cpr-nummer)'].strip():
            new_row = {
                'cpr': row['Leder 2 (cpr-nummer)'].replace('-', ''),
                'ansvar' : row['Lederansvar "Leder 2"'],
                'afdeling': row['SD kort navn (afd.kode)']
            }
            manager_rows.append(new_row)

importer.new_itsystem(
    identifier='AD',
    system_name='Active Directory'
)

sd = sd_importer.SdImport(
    importer,
    MUNICIPALTY_NAME,
    MUNICIPALTY_CODE,
    import_date_from=GLOBAL_GET_DATE,
    ad_info=ad_uuids,
    manager_rows=manager_rows,
)

for cpr, username in ad_usernames.items():
    if importer.check_if_exists('employee', cpr):
        importer.join_itsystem(
            employee=cpr,
            user_key=username,
            itsystem_ref='AD',
            date_from=None
        )

sd.employee_forced_uuids = ad_uuids
sd.employee_ad_usernames = ad_usernames

sd.create_ou_tree()
sd.create_employees()

importer.import_all()


"""
for info in sd.address_errors.values():
    print(info['DepartmentName'])
    print(info['DepartmentIdentifier'])
    print(info['PostalAddress']['StandardAddressIdentifier'])
    print(info['PostalAddress']['PostalCode'] + ' ' +
          info['PostalAddress']['DistrictName'])
    print()
    print()
print(len(sd.address_errors))
"""



from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import DateTime
from sqlalchemy import Integer
from sqlalchemy import String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.declarative import DeclarativeMeta

Base: DeclarativeMeta = declarative_base()


class Facet(Base):
    __tablename__ = "wfacetter"
    uuid = Column(String(36), nullable=False, primary_key=True)
    bvn = Column(String(250), nullable=False)


class Klasse(Base):
    __tablename__ = "wklasser"

    uuid = Column(String(36), nullable=False, primary_key=True)
    bvn = Column(String(250), nullable=False)
    titel = Column(String(250), nullable=False)
    facet_uuid = Column(String(36))  # , ForeignKey('facetter.uuid'))
    facet_bvn = Column(String(250), nullable=False)


class Bruger(Base):
    __tablename__ = "wbrugere"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    fornavn = Column(String(250))
    efternavn = Column(String(250))
    kaldenavn_fornavn = Column(String(250))
    kaldenavn_efternavn = Column(String(250))
    cpr = Column(String(250), nullable=True)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Enhed(Base):
    __tablename__ = "wenheder"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    navn = Column(String(250), nullable=False)
    bvn = Column(String(250), nullable=False)
    forældreenhed_uuid = Column(String(36), primary_key=False)
    enhedstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    enhedstype_titel = Column(String(250), nullable=False)
    enhedsniveau_uuid = Column(String(36))  # ForeignKey('klasser.uuid'))
    enhedsniveau_titel = Column(String(250))
    organisatorisk_sti = Column(String(1000))
    leder_uuid = Column(String(36))
    fungerende_leder_uuid = Column(String(36))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Adresse(Base):
    __tablename__ = "wadresser"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    værdi = Column(String(250))
    dar_uuid = Column(String(36))  # , ForeignKey('dar_adresser.uuid'))
    adressetype_uuid = Column(String(36))  # , ForeignKey("klasser.uuid"))
    adressetype_bvn = Column(String(250), nullable=False)
    adressetype_scope = Column(String(250), nullable=False)
    adressetype_titel = Column(String(250), nullable=False)
    synlighed_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    synlighed_scope = Column(String(250))
    synlighed_titel = Column(String(250))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Engagement(Base):
    __tablename__ = "wengagementer"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    bvn = Column(String(250), nullable=False)
    arbejdstidsfraktion = Column(Integer)
    engagementstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    engagementstype_titel = Column(String(250), nullable=False)
    primærtype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    primærtype_titel = Column(String(250))
    stillingsbetegnelse_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    stillingsbetegnelse_titel = Column(String(250), nullable=False)
    primær_boolean = Column(Boolean)
    udvidelse_1 = Column(String(250))
    udvidelse_2 = Column(String(250))
    udvidelse_3 = Column(String(250))
    udvidelse_4 = Column(String(250))
    udvidelse_5 = Column(String(250))
    udvidelse_6 = Column(String(250))
    udvidelse_7 = Column(String(250))
    udvidelse_8 = Column(String(250))
    udvidelse_9 = Column(String(250))
    udvidelse_10 = Column(String(250))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Rolle(Base):
    __tablename__ = "wroller"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    rolletype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    rolletype_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Tilknytning(Base):
    __tablename__ = "wtilknytninger"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    tilknytningstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    tilknytningstype_titel = Column(String(250))
    stillingsbetegnelse_uuid = Column(String(36))
    stillingsbetegnelse_titel = Column(String(250))
    it_forbindelse_uuid = Column(String(36))
    startdato = Column(String(10))
    slutdato = Column(String(10))
    primær_boolean = Column(Boolean, nullable=True)


class Orlov(Base):
    __tablename__ = "worlover"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bvn = Column(String(250), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    orlovstype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    orlovstype_titel = Column(String(250), nullable=False)
    engagement_uuid = Column(String(36))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class ItSystem(Base):
    __tablename__ = "wit_systemer"

    uuid = Column(String(36), nullable=False, primary_key=True)
    navn = Column(String(250), nullable=False)


class ItForbindelse(Base):
    __tablename__ = "wit_forbindelser"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    it_system_uuid = Column(String(36))  # , ForeignKey('it_systemer.uuid'))
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    brugernavn = Column(String(250))
    primær_boolean = Column(Boolean, nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Leder(Base):
    __tablename__ = "wledere"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    bruger_uuid = Column(String(36))  # , ForeignKey('brugere.uuid'))
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    ledertype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    ledertype_titel = Column(String(250), nullable=False)
    niveautype_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    niveautype_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class LederAnsvar(Base):
    __tablename__ = "wleder_ansvar"

    id = Column(Integer, nullable=False, primary_key=True)
    leder_uuid = Column(String(36))  # , ForeignKey('ledere.uuid'))
    lederansvar_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    lederansvar_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class KLE(Base):
    __tablename__ = "wkle"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    enhed_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    kle_aspekt_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    kle_aspekt_titel = Column(String(250), nullable=False)
    kle_nummer_uuid = Column(String(36))  # , ForeignKey('klasser.uuid'))
    kle_nummer_titel = Column(String(250), nullable=False)
    startdato = Column(String(10))
    slutdato = Column(String(10))


class Kvittering(Base):
    __tablename__ = "kvittering"

    id = Column(Integer, nullable=False, primary_key=True)
    query_tid = Column(DateTime)
    start_levering_tid = Column(DateTime)
    slut_levering_tid = Column(DateTime)


class Enhedssammenkobling(Base):
    __tablename__ = "wenhedssammenkobling"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    enhed1_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    enhed2_uuid = Column(String(36))  # , ForeignKey('enheder.uuid'))
    startdato = Column(String(10))
    slutdato = Column(String(10))


class DARAdresse(Base):
    __tablename__ = "wdar_adresser"

    id = Column(Integer, nullable=False, primary_key=True)
    uuid = Column(String(36), nullable=False)
    vejkode = Column(String(8))
    vejnavn = Column(String(250))
    husnr = Column(String(8))
    etage = Column(String(16))
    dør = Column(String(8))
    postnr = Column(String(8))
    postnrnavn = Column(String(250))
    kommunekode = Column(String(8))
    adgangsadresseid = Column(String(36))
    betegnelse = Column(String(250))
